# coding: utf-8
from __future__ import unicode_literals, division, absolute_import, print_function

from .version import __version__, __version_info__

__all__ = [
    '__version__',
    '__version_info__',
    'load_order',
]


def load_order():
    """
    Returns a list of the module and sub-module names for asn1crypto in
    dependency load order, for the sake of live reloading code

    :return:
        A list of unicode strings of module names, as they would appear in
        sys.modules, ordered by which module should be reloaded first
    """

    return [
        'asn1crypto._errors',
        'asn1crypto._int',
        'asn1crypto._ordereddict',
        'asn1crypto._teletex_codec',
        'asn1crypto._types',
        'asn1crypto._inet',
        'asn1crypto._iri',
        'asn1crypto.version',
        'asn1crypto.pem',
        'asn1crypto.util',
        'asn1crypto.parser',
        'asn1crypto.core',
        'asn1crypto.algos',
        'asn1crypto.keys',
        'asn1crypto.x509',
        'asn1crypto.crl',
        'asn1crypto.csr',
        'asn1crypto.ocsp',
        'asn1crypto.cms',
        'asn1crypto.pdf',
        'asn1crypto.pkcs12',
        'asn1crypto.tsp',
        'asn1crypto',
    ]



# ric.py - functions for handling Chinese Resident Identity Card Number
#
# Copyright (C) 2014 Jiangge Zhang
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""RIC No. (Chinese Resident Identity Card Number).

The RIC No. is the unique identifier for issued to China (PRC) residents.

The number consist of 18 digits in four sections. The first 6 digits refers to
the resident's location, followed by 8 digits representing the resident's birth
day in the form YYYY-MM-DD. The next 3 digits is the order code which is the
code used to disambiguate people with the same date of birth and address code.
Men are assigned to odd numbers, women assigned to even numbers. The final
digit is the checksum.

>>> validate('360426199101010071')
'360426199101010071'
"""

import datetime

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number).upper().strip()


def get_birth_date(number):
    """Split the date parts from the number and return the birth date.
    Note that in some cases it may return the registration date instead of
    the birth date and it may be a century off."""
    number = compact(number)
    year = int(number[6:10])
    month = int(number[10:12])
    day = int(number[12:14])
    try:
        return datetime.date(year, month, day)
    except ValueError:
        raise InvalidComponent()


def get_birth_place(number):
    """Use the number to look up the place of birth of the person."""
    from stdnum import numdb
    number = compact(number)
    results = numdb.get('cn/loc').info(number[:6])[0][1]
    if not results:
        raise InvalidComponent()
    return results


def calc_check_digit(number):
    """Calculate the check digit. The number passed should have the check
    digit included."""
    checksum = (1 - 2 * int(number[:-1], 13)) % 11
    return 'X' if checksum == 10 else str(checksum)


def validate(number):
    """Check if the number is a valid RIC number. This checks the length,
    formatting and birth date and place."""
    number = compact(number)
    if len(number) != 18:
        raise InvalidLength()
    if not isdigits(number[:-1]):
        raise InvalidFormat()
    if number[-1] != calc_check_digit(number):
        raise InvalidChecksum()
    get_birth_date(number)
    get_birth_place(number)
    return number


def is_valid(number):
    """Check if the number is a valid RIC number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



# -*- coding: utf-8 -*-

from datetime import datetime, timedelta

from dialogmanager import DialogManager


class CapptaCallback(object):
    def __init__(self, dialog_manager):
        # type: (DialogManager) -> None

        self.dialog_manager = dialog_manager
        self.last_response = dict(message=None, timestamp=None)

    def callback(self, pos_id, message, callback_type):
        try:
            if callback_type == "MENSAGEM":
                is_repeated = self._check_if_message_is_not_repeated(message)
                if is_repeated:
                    return

                self.dialog_manager.show_message(pos_id, message)

            elif callback_type == "REQUISITA_ACAO":
                split_message = message.split("\n")
                title = split_message[0]
                user_message = ""
                if len(split_message) > 1:
                    user_message = "\n".join(split_message[1:])

                resp = self.dialog_manager.show_numpad(pos_id, user_message, title)
                return "CANCELA" if not resp else str(resp)

        except BaseException as _:
            return "CANCELA"

    def _check_if_message_is_not_repeated(self, response):
        now = datetime.now()
        if response == self.last_response["message"] and (now - self.last_response["timestamp"]) < timedelta(seconds=3):
            return True
        else:
            self.last_response["message"] = response
            self.last_response["timestamp"] = now
            return False



from logging import Logger

from production.repository import ProductRepository
from typing import List, Optional, Union

from ._AddTagBox import AddTagBox
from ._ProductionBox import ProductionBox


class NoJitAddTagBox(AddTagBox):
    def __init__(self, name, sons, product_repository, tag, logger=None):
        # type: (str, Optional[Union[List[ProductionBox], ProductionBox]], ProductRepository, List[str], Logger) -> None
        part_codes = product_repository.get_no_jit_part_codes()
        super(NoJitAddTagBox, self).__init__(name, sons, tag, part_codes, logger)



# ubn.py - functions for handling Taiwanese UBN numbers
# coding: utf-8
#
# Copyright (C) 2020 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""UBN (Unified Business Number, 統一編號, Taiwanese tax number).

The Unified Business Number (UBN, 統一編號) is the number assigned to businesses
within Taiwan for tax (VAT) purposes. The number consists of 8 digits, the
last being a check digit.

More information:

* https://zh.wikipedia.org/wiki/統一編號
* https://findbiz.nat.gov.tw/fts/query/QueryBar/queryInit.do?request_locale=en

>>> validate('00501503')
'00501503'
>>> validate('00501502')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('12345')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format(' 0050150 3 ')
'00501503'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes surrounding
    whitespace.
    """
    return clean(number, ' -').strip()


def calc_checksum(number):
    """Calculate the checksum over the number."""
    # convert to numeric first, then sum individual digits
    weights = (1, 2, 1, 2, 1, 2, 4, 1)
    number = ''.join(str(w * int(n)) for w, n in zip(weights, number))
    return sum(int(n) for n in number) % 10


def validate(number):
    """Check if the number is a valid Taiwan UBN number.

    This checks the length, formatting and check digit.
    """
    number = compact(number)
    if len(number) != 8:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    checksum = calc_checksum(number)
    if not (checksum == 0 or (checksum == 9 and number[6] == '7')):
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid Taiwan UBN number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number)



import reportactions

from actions.config import (
    get_hide_unpriced_items_from_total_report,
    get_pre_account_with_discount,
    get_pre_account_allow_printing,
)
from typing import (
    Optional,
)
from sysactions import (
    action,
    get_model,
    get_current_order,
)


@action
def qsr_totaled_total_report(
    pos_id,                         # type: str
    is_mobile=False,                # type: Optional[bool]
):
    # type: (...) -> None

    model = get_model(pos_id)
    order = get_current_order(model)

    table_id = None
    print_pre_account_by = "qsr"
    order_is_totaled = order is not None and order.get("state") == "TOTALED"
    filter_seat = "0"

    reportactions.print_report(
        pos_id=pos_id,
        report_name="table_report",
        order=order,
        allow_printing=get_pre_account_allow_printing(),
        is_mobile=is_mobile.lower() == "true" if isinstance(is_mobile, basestring) else is_mobile,
        report_args=[
            table_id,
            order.get("orderId"),
            print_pre_account_by,
            get_hide_unpriced_items_from_total_report(),
            get_pre_account_with_discount() or order_is_totaled,
            filter_seat,
        ],
    )



from datetime import (
    datetime,
    timedelta,
)
from typing import Optional


class UserUpdateData(object):
    def __init__(
        self,
        user_id,            # type: int
        long_name,          # type: str
        password,           # type: str
        nfc_id_code,        # type: str
        level,              # type: int
        start_date,         # type: str
        end_date,           # type: str
        pay_rate,           # type: float
        status=None,        # type: Optional[int]
    ):
        # type: (...) -> None

        self.user_id = user_id
        self.long_name = long_name
        self.password = password
        self.nfc_id_code = nfc_id_code
        self.level = level
        self.start_date = start_date
        self.end_date = end_date
        self.pay_rate = pay_rate
        self.status = status

    def fill_user_status(self):
        start = datetime.strptime(self.start_date, "%Y-%m-%d")
        now = datetime.now()
        if self.end_date is not None:
            end = datetime.strptime(self.end_date, "%Y-%m-%d")
        else:
            one_day = timedelta(days=1)
            end = now + one_day

        active = False
        if start <= now <= end or int(self.level) >= 30:
            active = True

        self.status = "0" if active else "1"



from typing import (
    List,
    Dict,
)

from devicemanager.interactor.dto import DeviceStatusDto
from devicemanager.model import (
    DeviceConfig,
    DeviceStatus,
)
from devicemanager.repository import DeviceManagerRepository


class GetDevicesService(object):
    def __init__(
        self,
        repository,         # type: DeviceManagerRepository
        device_configs,     # type: List[DeviceConfig]
    ):
        # type: (...) -> None

        self.repository = repository
        self.available_names = []
        self.available_device_names = {}
        for config in device_configs:
            device_names_to_process = [config.default_device] + config.fallback_devices
            for device_name in device_names_to_process:
                if device_name not in self.available_device_names:
                    self.available_device_names[device_name] = True
                    self.available_names.append(device_name)

    def get_devices(
        self,   # type: GetDevicesService
    ):
        # type: (...) -> List[DeviceStatusDto]

        devices = self.repository.get_devices_with_last_history(device_names=self.available_names)
        processed_devices = dict(self.available_device_names)
        ret = self._create_in_database_devices(
            devices=devices,
            processed_devices=processed_devices,
        )
        ret.extend(self._create_configured_but_not_in_database(processed_devices=processed_devices))
        return sorted(ret, key=lambda x: x.name)

    @staticmethod
    def _create_in_database_devices(
        devices,            # type: List[DeviceStatus]
        processed_devices,  # type: Dict[str, bool]
    ):
        # type: (...) -> List[DeviceStatusDto]

        ret = []
        for device in devices:
            if device.device_name in processed_devices:
                ret.append(
                    DeviceStatusDto(
                        type="printer",
                        name=device.device_name,
                        last_update=device.history[0].timestamp,
                        last_update_operator=device.history[0].operator,
                        enabled=device.enabled,
                    )
                )
                del processed_devices[device.device_name]
        return ret

    @staticmethod
    def _create_configured_but_not_in_database(
        processed_devices
    ):
        # type: (...) -> List[DeviceStatusDto]

        ret = []
        for key in list(processed_devices.keys()):
            ret.append(
                DeviceStatusDto(
                    type="printer",
                    name=key,
                    last_update=None,
                    last_update_operator=None,
                    enabled=True,
                )
            )
        return ret



import pathlib
import functools

from typing import Dict, Union


####
# from jaraco.path 3.4.1

FilesSpec = Dict[str, Union[str, bytes, 'FilesSpec']]  # type: ignore


def build(spec: FilesSpec, prefix=pathlib.Path()):
    """
    Build a set of files/directories, as described by the spec.

    Each key represents a pathname, and the value represents
    the content. Content may be a nested directory.

    >>> spec = {
    ...     'README.txt': "A README file",
    ...     "foo": {
    ...         "__init__.py": "",
    ...         "bar": {
    ...             "__init__.py": "",
    ...         },
    ...         "baz.py": "# Some code",
    ...     }
    ... }
    >>> target = getfixture('tmp_path')
    >>> build(spec, target)
    >>> target.joinpath('foo/baz.py').read_text(encoding='utf-8')
    '# Some code'
    """
    for name, contents in spec.items():
        create(contents, pathlib.Path(prefix) / name)


@functools.singledispatch
def create(content: Union[str, bytes, FilesSpec], path):
    path.mkdir(exist_ok=True)
    build(content, prefix=path)  # type: ignore


@create.register
def _(content: bytes, path):
    path.write_bytes(content)


@create.register
def _(content: str, path):
    path.write_text(content, encoding='utf-8')


# end from jaraco.path
####



from application.repository import I18nRepository
from sysactions import send_message, get_model
from msgbus import FM_PARAM, TK_I18N_GETMSG, TK_SYS_ACK


class MwI18nRepository(I18nRepository):
    def __init__(self, pos_id):
        self.pos_id = pos_id

        self.lang = self._get_language(pos_id)

        l10n_params = [
            "L10_PYTHON_DATE_FORMAT",
            "L10_PYTHON_DATETIME_FORMAT",
            "L10N_DECIMAL_PLACES",
            "L10N_CURRENCY_DECIMALS",
            "L10N_CURRENCY_SYMBOL",
            "L10N_DECIMALS_SEPARATOR",
            "L10N_THOUSANDS_SEPARATOR",
            "L10N_CURRENCY_SYMBOL_POSITION"
        ]
        l10n_param = self.lang + "\0"
        l10n_param += "\0".join(l10n_params)

        msg = send_message("I18N", TK_I18N_GETMSG, FM_PARAM, l10n_param)
        if msg.token != TK_SYS_ACK:
            raise Exception("Invalid token from I18n: {} - {}".format(msg.token, msg.data))
        l10n_data = self.remove_language_and_last_zero(msg.data)
        self.l10n_params = {}
        count = 0
        for data in l10n_data:
            self.l10n_params[l10n_params[count]] = data
            count += 1

    def get_date_format(self):
        return self.l10n_params["L10_PYTHON_DATE_FORMAT"]

    def get_date_time_format(self):
        return self.l10n_params["L10_PYTHON_DATETIME_FORMAT"]

    def get_decimal_separator(self):
        return self.l10n_params["L10N_DECIMALS_SEPARATOR"]

    def get_thousands_separator(self):
        return self.l10n_params["L10N_THOUSANDS_SEPARATOR"]

    def get_currency_symbol(self):
        return self.l10n_params["L10N_CURRENCY_SYMBOL"]

    def get_currency_symbol_position(self):
        return self.l10n_params["L10N_CURRENCY_SYMBOL_POSITION"]

    def get_decimal_places(self):
        return self.l10n_params["L10N_DECIMAL_PLACES"]

    def internationalize(self, labels):
        dict_words_translated = {}

        param = self.lang + "\0"
        for label in labels:
            if label.startswith("$"):
                param += label[1:] + "\0"
        param = param[:-1]

        msg = send_message("I18N", TK_I18N_GETMSG, FM_PARAM, param)
        if msg.token != TK_SYS_ACK:
            raise Exception("Invalid token from I18N: {} - {}".format(msg.token, msg.data))
        translated_words = self.remove_language_and_last_zero(msg.data)

        count = 0
        for label in labels:
            dict_words_translated[label] = translated_words[count].decode("utf-8")
            count += 1

        return dict_words_translated

    @staticmethod
    def _get_language(pos_id):
        # type: (int) -> str
        model = get_model(pos_id)
        lang = model.findall('Language')[0].get('name')
        return lang

    @staticmethod
    def remove_language_and_last_zero(data):
        return data.split('\0')[1:-1]



from logging import Logger
from typing import Dict

from production.box import RemoveReprocessedOrdersBox
from production.command import ProductionCommandProcessor
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class RemoveReprocessedOrdersBoxBuilder(BoxBuilder):
    def __init__(
        self,
        command_processor,  # type: ProductionCommandProcessor
        loggers,            # type: Dict[unicode, Logger]
    ):
        # type: (...) -> None

        super(RemoveReprocessedOrdersBoxBuilder, self).__init__(loggers)

        self.command_processor = command_processor

    def build(self, config):
        # type: (BoxConfiguration) -> RemoveReprocessedOrdersBox
        
        return RemoveReprocessedOrdersBox(
            name=config.name,
            sons=config.sons,
            command_processor=self.command_processor,
            reprocessed_time_limit=self.get_extra(config, "ReprocessedTimeLimit", None),
            logger=self.get_logger(config),
        )



import logging
from datetime import datetime

from typing import List

from actions import mb_context
from application.repository import UserRepository
from utilfunctions import get_seconds_to_renew_cache, reached_seconds_to_renew_cache

own_delivery_men = []
last_cache_time = None


class UserInteractor(object):
    def __init__(
            self
    ):
        # type: (...) -> None

        self.logger = logging.getLogger("PosActions")
        self.user_repository = UserRepository(mb_context)

    def get_own_delivery_men(
            self
    ):
        # type: (...) -> List

        global own_delivery_men
        global last_cache_time

        try:
            seconds_to_renew = get_seconds_to_renew_cache(60 * 5)
            if not own_delivery_men or reached_seconds_to_renew_cache(last_cache_time, seconds_to_renew):
                own_delivery_men = self.user_repository.get_own_delivery_men()
                last_cache_time = datetime.now()

            return [delivery_men.get("name", "") for delivery_men in own_delivery_men]
        except Exception as _:
            self.logger.exception("Error getting own delivers")
            return []



from typing import Optional

from ._PaymentResponse import PaymentResponse


class PaymentResponseData(PaymentResponse):
    def __init__(
        self,
        media,                      # type: str
        result,                     # type: str
        card_number,                # type: str
        auth_code,                  # type: str
        approved_amount,            # type: str
        id_auth,                    # type: str
        cnpj_auth,                  # type: str
        bandeira,                   # type: str
        receipt_merchant,           # type: str
        receipt_customer,           # type: str
        expiration_date,            # type: str
        nsu,                        # type: str
        transaction_processor,      # type: str
        last_digits,                # type: str
        sat_acquirer,               # type: str
        t_pag,                      # type: str
        t_band,                     # type: str
        nsu_host="",                # type: Optional[str]
        electronic_type="",         # type: Optional[str]
        transaction_identifier=""   # type: Optional[str]
    ):
        # type: (...) -> None

        self.media = media
        self.result = result
        self.card_number = card_number
        self.auth_code = auth_code
        self.approved_amount = approved_amount
        self.id_auth = id_auth
        self.cnpj_auth = cnpj_auth
        self.bandeira = bandeira
        self.receipt_merchant = receipt_merchant
        self.receipt_customer = receipt_customer
        self.expiration_date = expiration_date
        self.nsu = nsu
        self.transaction_processor = transaction_processor
        self.last_digits = last_digits
        self.sat_acquirer = sat_acquirer
        self.t_pag = t_pag
        self.t_band = t_band
        self.nsu_host = nsu_host
        self.electronic_type = electronic_type
        self.transaction_identifier = transaction_identifier

    def to_xml(self):
        return """<EftXml 
            Media='%s'
            Result='%s'
            CardNumber='%sXXXXXXXXXX'
            AuthCode='%s'
            ApprovedAmount='%s'
            IdAuth='%s'
            CNPJAuth='%s'
            Bandeira='%s'
            ReceiptMerchant='%s'
            ReceiptCustomer='%s'
            ExpirationDate='%s'
            Nsu='%s'
            TransactionProcessor='%s'
            LastDigits='%s'
            AccreditorSat='%s'
            tPag='%s'
            tBand='%s'
            NsuHost='%s'
            ElectronicType='%s'
            TransactionIdentifier='%s'
            />
            """ % (
            self.media,
            self.result,
            self.card_number,
            self.auth_code,
            self.approved_amount,
            self.id_auth,
            self.cnpj_auth,
            self.bandeira,
            self.receipt_merchant,
            self.receipt_customer,
            self.expiration_date,
            self.nsu,
            self.transaction_processor,
            self.last_digits,
            self.sat_acquirer,
            self.t_pag,
            self.t_band,
            self.nsu_host,
            self.electronic_type,
            self.transaction_identifier
        )



from __future__ import annotations

import abc
import typing
from typing import Any
from typing import ClassVar

if typing.TYPE_CHECKING:
    from .interfaces import DBAPIConnection
    from .interfaces import Dialect


class ConnectionCharacteristic(abc.ABC):
    """An abstract base for an object that can set, get and reset a
    per-connection characteristic, typically one that gets reset when the
    connection is returned to the connection pool.

    transaction isolation is the canonical example, and the
    ``IsolationLevelCharacteristic`` implementation provides this for the
    ``DefaultDialect``.

    The ``ConnectionCharacteristic`` class should call upon the ``Dialect`` for
    the implementation of each method.   The object exists strictly to serve as
    a dialect visitor that can be placed into the
    ``DefaultDialect.connection_characteristics`` dictionary where it will take
    effect for calls to :meth:`_engine.Connection.execution_options` and
    related APIs.

    .. versionadded:: 1.4

    """

    __slots__ = ()

    transactional: ClassVar[bool] = False

    @abc.abstractmethod
    def reset_characteristic(
        self, dialect: Dialect, dbapi_conn: DBAPIConnection
    ) -> None:
        """Reset the characteristic on the connection to its default value."""

    @abc.abstractmethod
    def set_characteristic(
        self, dialect: Dialect, dbapi_conn: DBAPIConnection, value: Any
    ) -> None:
        """set characteristic on the connection to a given value."""

    @abc.abstractmethod
    def get_characteristic(
        self, dialect: Dialect, dbapi_conn: DBAPIConnection
    ) -> Any:
        """Given a DBAPI connection, get the current value of the
        characteristic.

        """


class IsolationLevelCharacteristic(ConnectionCharacteristic):
    transactional: ClassVar[bool] = True

    def reset_characteristic(
        self, dialect: Dialect, dbapi_conn: DBAPIConnection
    ) -> None:
        dialect.reset_isolation_level(dbapi_conn)

    def set_characteristic(
        self, dialect: Dialect, dbapi_conn: DBAPIConnection, value: Any
    ) -> None:
        dialect._assert_and_set_isolation_level(dbapi_conn, value)

    def get_characteristic(
        self, dialect: Dialect, dbapi_conn: DBAPIConnection
    ) -> Any:
        return dialect.get_isolation_level(dbapi_conn)



# -*- coding: utf-8 -*-

import base64
import time

from nfcebuilder.nfceutil import NfeSigner, XmlEnveloper
from nfcecanceler import NfceCancelerParameters


class NfceCancelerRequestBuilder(object):
    NAMESPACE_NFE = "http://www.portalfiscal.inf.br/nfe"
    NAMESPACE_WEB_SERVICE = "http://www.portalfiscal.inf.br/nfe/wsdl/RecepcaoEvento"
    NAMESPACE_WEB_SERVICE_4 = "http://www.portalfiscal.inf.br/nfe/wsdl/NFeRecepcaoEvento4"

    def __init__(self, canceler_parameters, nfce_signer, xml_enveloper):
        # type: (NfceCancelerParameters, NfeSigner, XmlEnveloper) -> None
        self.canceler_parameters = canceler_parameters
        self.nfce_signer = nfce_signer
        self.xml_enveloper = xml_enveloper

    def build_request(self, order_response, justificativa, versao_ws):
        xml_response = base64.b64decode(order_response + "===")

        idx_start = xml_response.find("<chNFe>") + 7
        idx_final = xml_response.find("</chNFe>")
        chave = xml_response[idx_start:idx_final]

        idx_start = xml_response.find("<nProt>") + 7
        idx_final = xml_response.find("</nProt>")
        nprot = xml_response[idx_start:idx_final]

        parameters_xml = self.canceler_parameters.fill_parameters(chave, nprot, justificativa)

        idlote = int(round(time.time() * 1000))
        cancel_pattern = "<envEvento xmlns=\"{0:s}\" versao=\"1.00\"><idLote>%d</idLote>{1:s}</envEvento>" % idlote

        xml = cancel_pattern.format(NfceCancelerRequestBuilder.NAMESPACE_NFE, parameters_xml)

        signed_xml = self.nfce_signer.sign_cancel_xml(xml)

        envelope = self.xml_enveloper.envelop(signed_xml,
                                              NfceCancelerRequestBuilder.NAMESPACE_WEB_SERVICE if versao_ws in (1, 3)
                                              else NfceCancelerRequestBuilder.NAMESPACE_WEB_SERVICE_4, versao_ws)

        return envelope



# nir.py - functions for handling French NIR numbers
# coding: utf-8
#
# Copyright (C) 2016 Dimitri Papadopoulos
# Copyright (C) 2016-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NIR (French personal identification number).

The NIR (Numero d'Inscription au Repertoire national d'identification des
personnes physiques) is used to identify persons in France. It is popularly
known as the "social security number" and sometimes referred to as an INSEE
number. All persons born in France are registered in the Repertoire national
d'identification des personnes physiques (RNIPP) and assigned a NIR.

The number consists of 15 digits: the first digit indicates the gender,
followed by 2 digits for the year or birth, 2 for the month of birth, 5 for
the location of birth (COG), 3 for a serial and 2 check digits.

More information:

* https://www.insee.fr/en/metadonnees/definition/c1409
* https://en.wikipedia.org/wiki/INSEE_code
* http://resoo.org/docs/_docs/regles-numero-insee.pdf
* https://fr.wikipedia.org/wiki/Numéro_de_sécurité_sociale_en_France
* https://xml.insee.fr/schema/nir.html

>>> validate('2 95 10 99 126 111 93')
'295109912611193'
>>> validate('295109912611199')  # invalid check digit
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> validate('253072B07300470')
'253072B07300470'
>>> validate('253072A07300443')
'253072A07300443'
>>> validate('253072C07300443')
Traceback (most recent call last):
    ...
InvalidFormat: ...
>>> validate('6546546546546703')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('295109912611193')
'2 95 10 99 126 111 93'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' .').strip().upper()


def calc_check_digits(number):
    """Calculate the check digits for the number."""
    department = number[5:7]
    if department == '2A':
        number = number[:5] + '19' + number[7:]
    elif department == '2B':
        number = number[:5] + '18' + number[7:]
    return '%02d' % (97 - (int(number[:13]) % 97))


def validate(number):
    """Check if the number provided is valid. This checks the length
    and check digits."""
    number = compact(number)
    if not isdigits(number[:5]) or not isdigits(number[7:]):
        raise InvalidFormat()
    if not isdigits(number[5:7]) and number[5:7] not in ('2A', '2B'):
        raise InvalidFormat()
    if len(number) != 15:
        raise InvalidLength()
    if calc_check_digits(number) != number[13:]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is valid."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number, separator=' '):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return separator.join((
        number[:1], number[1:3], number[3:5], number[5:7], number[7:10],
        number[10:13], number[13:]))



from ._OrderChangerProductionBox import OrderChangerProductionBox


class ProdStateTimerBox(OrderChangerProductionBox):

    def __init__(self, name, sons, prod_state, logger=None):
        super(ProdStateTimerBox, self).__init__(name, sons, logger)
        self.prod_state = prod_state

    def change_order(self, order):
        order_timestamp = self.get_order_timestamp_by_prod_state(order)
        if order_timestamp:
            order.display_time = order_timestamp

        return order

    def get_order_timestamp_by_prod_state(self, order):
        for state in reversed(order.prod_state_history):
            if state.prod_state == self.prod_state:
                return state.timestamp
        return None



"""
An implementation of the basestring type for Python 3

Example use:

>>> s = b'abc'
>>> assert isinstance(s, basestring)
>>> from past.types import str as oldstr
>>> s2 = oldstr(b'abc')
>>> assert isinstance(s2, basestring)

"""

import sys

from past.utils import with_metaclass, PY2

if PY2:
    str = unicode

ver = sys.version_info[:2]


class BaseBaseString(type):
    def __instancecheck__(cls, instance):
        return isinstance(instance, (bytes, str))

    def __subclasshook__(cls, thing):
        # TODO: What should go here?
        raise NotImplemented


class basestring(with_metaclass(BaseBaseString)):
    """
    A minimal backport of the Python 2 basestring type to Py3
    """


__all__ = ['basestring']




from datetime import datetime
import os
from logging import Logger

import cfgtools
import logging

import persistence
import sysactions

from helper import import_pydevd, config_logger
from msgbus import MBEasyContext
from systools import sys_parsefilepath

from messagehandlerbuilder import ProductionMessageHandlerBuilder
from messagehandlerbuilder.token import all_events
from mbcontextmessagehandler import MbContextMessageBus, MbContextMessageHandler
from production.model import NoBoxesConfiguredException, NoViewsConfiguredException, PicklistCleanerManagerConfig
from production.util import ProductionConfig
from production.treebuilder import ViewConfiguration, BoxConfiguration, TreeConfig
from typing import List, Dict


def main():
    import_pydevd(os.environ["LOADERCFG"], 9131, False)
    config_logger(os.environ["LOADERCFG"], "ProductionPerformanceEvent", max_files=5, max_size=12582912)
    config_logger(os.environ["LOADERCFG"], "ProductionPerformanceCommand", max_files=5, max_size=12582912)
    config_logger(os.environ["LOADERCFG"], "ProductionPrinter", max_files=5, max_size=12582912)
    config_logger(os.environ["LOADERCFG"], "PicklistDatabaseManager", max_files=5, max_size=12582912)

    config = cfgtools.read(os.environ["LOADERCFG"])
    max_files = int(config.find_value("Production.Manager.MaxFileLogs", 5))
    config_logger(os.environ["LOADERCFG"], "Production", max_files=max_files, max_size=12582912, add_error_handler=True)

    base_logger = logging.getLogger("Production")
    debug_logger = build_logger(base_logger, "ProductionDebug", logging.DEBUG)
    info_logger = build_logger(base_logger, "ProductionInfo", logging.INFO)
    warning_logger = build_logger(base_logger, "ProductionWarning", logging.WARNING)
    error_logger = build_logger(base_logger, "ProductionError", logging.ERROR)

    loggers = {
        u"DEBUG": debug_logger,
        u"INFO": info_logger,
        u"WARNING": warning_logger,
        u"ERROR": error_logger,
    }

    boxes_config = parse_box_configurations(config, "Production.Boxes")
    views_config = parse_view_configurations(config)
    branches_config = parse_box_configurations(config, "Production.Branches")
    clean_picklist_database_service_config = get_clean_picklist_database_service_config(config=config)

    if len(boxes_config) == 0:
        raise NoBoxesConfiguredException("No boxes configured")
    if len(views_config) == 0:
        raise NoViewsConfiguredException("No views configured")

    purge_interval = int(config.find_value("Production.PurgeInterval", "1"))
    orders_life_time = int(config.find_value("Production.OrdersLifeTime", "20"))
    manager_log_level = config.find_value("Production.Manager.LogLevel", "ERROR")
    audit_logger_prod_states = config.find_values("Production.Manager.AuditLogProdStates") or []
    kds_manager = config.find_value("Production.KdsManager", "NKDSCTRL")
    timeout_to_cleanup_devices = int(config.find_value("Production.TimeToCleanDevicesCache") or 5)

    mb_context = MBEasyContext("NewProduction")

    sysactions.mbcontext = mb_context

    message_bus = MbContextMessageBus(mb_context)
    production_config = ProductionConfig(
        boxes_config=boxes_config,
        branches_config=branches_config,
        views_config=views_config,
        picklist_database_manager_config=clean_picklist_database_service_config,
    )

    production_connection_string = _get_connection_string(
        config=config,
        db_name="production"
    )

    product_connection_string = _get_connection_string(
        config=config,
        db_name="product"
    )
    message_handler_builder = ProductionMessageHandlerBuilder(
        message_bus=message_bus,
        purge_interval=purge_interval,
        orders_life_time=orders_life_time,
        manager_log_level=manager_log_level,
        audit_logger_prod_states=audit_logger_prod_states,
        kds_manager=kds_manager,
        production_config=production_config,
        loggers=loggers,
        timeout_to_cleanup_devices=timeout_to_cleanup_devices,
        production_connection_string=production_connection_string,
        product_connection_string=product_connection_string,
    )

    message_handler = MbContextMessageHandler(
        message_bus=message_bus,
        service_name="ProductionSystem",
        service_type="Production",
        required_services="StoreWideConfig|NKDSCTRL|Persistence",
        event_handler_builder=message_handler_builder,
    )

    message_handler.subscribe_reentrant_events(
        subjects=all_events()
    )

    _wait_persistence(
        message_bus_context=mb_context,
    )

    return message_handler.handle_events()


def parse_view_configurations(config):
    # type: (cfgtools.Group) -> List[ViewConfiguration]
    ret = []
    views_group = config.find_group("Production.Views")  # type: cfgtools.Group
    for group in views_group.groups:
        name = group.name
        if name is None:
            raise Exception("View without name")

        view_type = None
        log_level = "ERROR"
        extra_config = {}
        for key in group.keys:
            if key.name == "Type":
                view_type = key.value()
            elif key.name == "LogLevel":
                log_level = key.value()
            else:
                extra_config[key.name] = key.values if len(key.values) > 1 else key.value()

        if view_type is None:
            raise Exception("View without type: {}".format(name))

        ret.append(ViewConfiguration(name, view_type, log_level, extra_config))

    return ret


def parse_box_configurations(config, config_path):
    # type: (cfgtools.Group, str) -> List[TreeConfig]
    ret = []
    boxes_group = config.find_group(config_path)  # type: cfgtools.Group
    if not boxes_group:
        return ret

    for tree in boxes_group.groups:  # type: cfgtools.Group
        tree_boxes = []
        branch_name = tree.name
        for group in tree.groups:
            name = group.name.strip()
            if name is None:
                raise Exception("View without name")

            box_type = None
            log_level = "ERROR"
            extra_config = {}
            sons = []
            for key in group.keys:
                if key.name == "Type":
                    box_type = key.value().strip()
                elif key.name == "LogLevel":
                    log_level = key.value().strip()
                elif key.name == "Sons":
                    if len(key.values) > 0:
                        sons.extend([x.strip() for x in key.values])
                    else:
                        sons.append(key.value().strip())
                else:
                    extra_config[key.name] = key.values if len(key.values) > 1 else key.value()

            for inner_group in group.groups:
                extra_config[inner_group.name] = get_extra_config(inner_group)

            if box_type is None:
                raise Exception("View without type: {}".format(name))

            tree_boxes.append(BoxConfiguration(name, box_type, sons, log_level, extra_config))
        ret.append(TreeConfig(branch_name, tree_boxes))

    return ret


def get_clean_picklist_database_service_config(
    config,     # type: cfgtools.Group
):
    # type: (...) -> PicklistCleanerManagerConfig

    start_time = config.find_value("Production.Maintenance.Window.StartTime") or None
    end_time = config.find_value("Production.Maintenance.Window.EndTime") or None
    if start_time and end_time:
        start_time = datetime.strptime(start_time, "%H:%M:%S").time()
        end_time = datetime.strptime(end_time, "%H:%M:%S").time()

    return PicklistCleanerManagerConfig(
        start_time=start_time,
        end_time=end_time,
        days_to_keep_data=int(config.find_value("Production.DaysToKeepData", 45))
    )


def get_extra_config(group):
    # type: (cfgtools.Group) -> Dict[str, str]
    inner_dict = {}
    for key in group.keys:
        inner_dict[key.name] = key.values if len(key.values) > 1 else key.value()
    for inner_group in group.groups:
        inner_dict[inner_group.name] = get_extra_config(inner_group)
    return inner_dict


def build_logger(base_logger, name, level):
    # type: (Logger, str, int) -> Logger
    logger = logging.getLogger(name)
    for handler in base_logger.handlers:
        logger.addHandler(handler)
    logger.setLevel(level)

    return logger


def _get_connection_string(
    config,     # type: cfgtools.Group
    db_name,    # type: str
):
    # type: (...) -> str

    default_value = os.path.join("{HVDATADIR}", "databases")
    path = config.find_value("Production.DatabaseDir") or default_value
    databases_dir = sys_parsefilepath(
        path=path,
    )
    return os.path.join(databases_dir, "{}.db".format(db_name))


def _wait_persistence(
    message_bus_context,    # type: MBEasyContext
):
    # type: (...) -> None

    persistence.Driver().open(
        mbcontext=message_bus_context,
    ).close()



from typing import ContextManager

from .repository import EventRepository


class EventRepositoryFactory(object):
    def build(self) -> ContextManager[EventRepository]:
        raise NotImplementedError()



from abc import ABCMeta, abstractmethod

from messageprocessor import MessageProcessor


class MessageMessageProcessorBuilder(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def build_message_processor(
        self,
        token,  # type: int
    ):
        # type: (...) -> MessageProcessor
        raise NotImplementedError()



from old_helper import round_half_away_from_zero

from typing import Dict, Any
from pos_model import Order
from _NfceXmlPartBuilder import NfceXmlPartBuilder
from _ContextKeys import ContextKeys


class TotalBuilder(NfceXmlPartBuilder):
    def __init__(self, versao_ws):
        # type: (int) -> None
        self.versao_ws = versao_ws

    def build_xml(self, order, context):
        # type: (Order, Dict[unicode, Any]) -> unicode
        v_prod = context[ContextKeys.total_prod]
        v_desc = 0
        v_other = 0
        for sale_line in order.sale_items:
            item_discount = float(sale_line.item_discount or 0)
            if item_discount > 0:
                v_desc += item_discount
            else:
                v_other += abs(item_discount)

        v_desc = round_half_away_from_zero(
            number=v_desc,
            precision=2,
        )
        v_other = round_half_away_from_zero(
            number=v_other,
            precision=2,
        )

        discount_value = float(order.discount_amount or 0)
        context[ContextKeys.v_nf] = v_nf = v_prod - discount_value

        total_icsm = context[ContextKeys.total_icms]
        total_pis = context[ContextKeys.total_pis]
        total_cofins = context[ContextKeys.total_cofins]
        v_tot_trib = total_icsm + total_pis + total_cofins

        xml = "<total><ICMSTot>"
        xml += "<vBC>{:.02f}</vBC>".format(context[ContextKeys.total_icms_vbc])
        xml += "<vICMS>{:.02f}</vICMS>".format(total_icsm)
        xml += "<vICMSDeson>{:.02f}</vICMSDeson>".format(context.get(ContextKeys.deson, 0))

        if self.versao_ws not in (1, 3):
            xml += "<vFCP>{:.02f}</vFCP>".format(context.get(ContextKeys.total_fcp, 0))

        xml += "<vBCST>0.00</vBCST><vST>0.00</vST>"

        if self.versao_ws not in (1, 3):
            xml += "<vFCPST>{:.02f}</vFCPST>".format(context.get(ContextKeys.total_fcp_st, 0))
            xml += "<vFCPSTRet>{:.02f}</vFCPSTRet>".format(context.get(ContextKeys.total_fcp_ant, 0))

        xml += "<vProd>{:.02f}</vProd>".format(v_prod)
        xml += "<vFrete>0.00</vFrete><vSeg>0.00</vSeg>"
        xml += "<vDesc>{:.02f}</vDesc>".format(v_desc)
        xml += "<vII>0.00</vII><vIPI>0.00</vIPI>"

        if self.versao_ws not in (1, 3):
            xml += "<vIPIDevol>0.00</vIPIDevol>"

        xml += "<vPIS>{:.02f}</vPIS>".format(total_pis)
        xml += "<vCOFINS>{:.02f}</vCOFINS>".format(total_cofins)
        xml += "<vOutro>{:.02f}</vOutro>".format(v_other)
        xml += "<vNF>{:.02f}</vNF>".format(v_nf)
        xml += "<vTotTrib>{:.02f}</vTotTrib>".format(v_tot_trib)
        xml += "</ICMSTot></total>"

        return xml



from setuptools import setup, find_packages

setup(
    name="EventManagerApi",
    version="1.0.0",
    packages=find_packages(where="."),
    install_requires=[],
    url="https://www.e-deploy.com.br",
    author="Eduardo Almeida",
    author_email="eduardo.almeida@e-deploy.com.br"
)



from contextlib import suppress
from io import TextIOWrapper

from . import abc


class SpecLoaderAdapter:
    """
    Adapt a package spec to adapt the underlying loader.
    """

    def __init__(self, spec, adapter=lambda spec: spec.loader):
        self.spec = spec
        self.loader = adapter(spec)

    def __getattr__(self, name):
        return getattr(self.spec, name)


class TraversableResourcesLoader:
    """
    Adapt a loader to provide TraversableResources.
    """

    def __init__(self, spec):
        self.spec = spec

    def get_resource_reader(self, name):
        return CompatibilityFiles(self.spec)._native()


def _io_wrapper(file, mode='r', *args, **kwargs):
    if mode == 'r':
        return TextIOWrapper(file, *args, **kwargs)
    elif mode == 'rb':
        return file
    raise ValueError(f"Invalid mode value '{mode}', only 'r' and 'rb' are supported")


class CompatibilityFiles:
    """
    Adapter for an existing or non-existent resource reader
    to provide a compatibility .files().
    """

    class SpecPath(abc.Traversable):
        """
        Path tied to a module spec.
        Can be read and exposes the resource reader children.
        """

        def __init__(self, spec, reader):
            self._spec = spec
            self._reader = reader

        def iterdir(self):
            if not self._reader:
                return iter(())
            return iter(
                CompatibilityFiles.ChildPath(self._reader, path)
                for path in self._reader.contents()
            )

        def is_file(self):
            return False

        is_dir = is_file

        def joinpath(self, other):
            if not self._reader:
                return CompatibilityFiles.OrphanPath(other)
            return CompatibilityFiles.ChildPath(self._reader, other)

        @property
        def name(self):
            return self._spec.name

        def open(self, mode='r', *args, **kwargs):
            return _io_wrapper(self._reader.open_resource(None), mode, *args, **kwargs)

    class ChildPath(abc.Traversable):
        """
        Path tied to a resource reader child.
        Can be read but doesn't expose any meaningful children.
        """

        def __init__(self, reader, name):
            self._reader = reader
            self._name = name

        def iterdir(self):
            return iter(())

        def is_file(self):
            return self._reader.is_resource(self.name)

        def is_dir(self):
            return not self.is_file()

        def joinpath(self, other):
            return CompatibilityFiles.OrphanPath(self.name, other)

        @property
        def name(self):
            return self._name

        def open(self, mode='r', *args, **kwargs):
            return _io_wrapper(
                self._reader.open_resource(self.name), mode, *args, **kwargs
            )

    class OrphanPath(abc.Traversable):
        """
        Orphan path, not tied to a module spec or resource reader.
        Can't be read and doesn't expose any meaningful children.
        """

        def __init__(self, *path_parts):
            if len(path_parts) < 1:
                raise ValueError('Need at least one path part to construct a path')
            self._path = path_parts

        def iterdir(self):
            return iter(())

        def is_file(self):
            return False

        is_dir = is_file

        def joinpath(self, other):
            return CompatibilityFiles.OrphanPath(*self._path, other)

        @property
        def name(self):
            return self._path[-1]

        def open(self, mode='r', *args, **kwargs):
            raise FileNotFoundError("Can't open orphan path")

    def __init__(self, spec):
        self.spec = spec

    @property
    def _reader(self):
        with suppress(AttributeError):
            return self.spec.loader.get_resource_reader(self.spec.name)

    def _native(self):
        """
        Return the native reader if it supports files().
        """
        reader = self._reader
        return reader if hasattr(reader, 'files') else self

    def __getattr__(self, attr):
        return getattr(self._reader, attr)

    def files(self):
        return CompatibilityFiles.SpecPath(self.spec, self._reader)


def wrap_spec(package):
    """
    Construct a package spec with traversable compatibility
    on the spec/loader/reader.
    """
    return SpecLoaderAdapter(package.__spec__, TraversableResourcesLoader)



from messagebus import (
    AckMessage,
    DataType,
    NakMessage
)

from ._CommonProcessor import _CommonProcessor
from model import TestConnectionData


class TestConnectionProcessor(_CommonProcessor):
    def parse_data(self, data, event_type):
        return TestConnectionData(*data.split(';'))

    def call_business(self, test_connection_data):
        return self.payment_router.get_connection_status(test_connection_data)

    def format_response(self, message_bus, message, event, input_obj, result):
        if result:
            message_bus.reply_message(message, AckMessage(data=str(result), data_type=DataType.string))
        else:
            message_bus.reply_message(message, NakMessage())



from ._AwakableSleeper import AwakableSleeper
from ._Clock import Clock
from ._ConditionSleeper import ConditionSleeper
from ._RealClock import RealClock
from ._RealSleeper import RealSleeper
from ._Sleeper import Sleeper



from typing import (
    Any,
    Optional,
)

from messagebus import (
    Event,
    Message,
    MessageBus,
)
from messageprocessorutil import UuidMessageProcessor

from dailygoals_services.interactor.update_total_sold import UpdateTotalSoldTableInteractor
from externals.util import LoggerUtil

logger = LoggerUtil.get_logger_name(__name__)


class UpdateTotalSoldTableProcessor(UuidMessageProcessor):
    def __init__(self, interactor):
        # type: (UpdateTotalSoldTableInteractor) -> None

        super(UpdateTotalSoldTableProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(self, data, event_type):
        # type: (Any, str) -> Any

        return data

    def call_business(self, params):
        # type: (Any) -> None

        self.interactor.update(params)

    def format_response(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Event]
        request,        # type: Any
        response,       # type: Any
    ):
        # type: (...) -> None

        logger.info("Total sold table updated for request")

    def format_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Event]
        request,        # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        exception_name = str(type(exception).__name__)
        logger.info("Error updating total sold table from request: {}".format(exception_name))

    def format_parse_exception(
        self,
        message_bus,    # type: MessageBus
        message,        # type: Message
        event,          # type: Optional[Event]
        request,        # type: Any
        exception,      # type: Exception
    ):
        # type: (...) -> None

        logger.info("Error getting parsing request: {}".format(str(exception)))



# -*- coding: utf-8 -*-

import logging
import os
import xml.etree.ElementTree as eTree

import cfgtools
import pyscripts
import sysactions
from cfgtools import Configuration
from helper import import_pydevd, remove_accents
from messagehandler import MessageHandler
from modelcommon.model import Store
from msgbus import MBEasyContext
from old_helper import OrderTaker, PosUtil, config_logger, read_swconfig, remove_punctuation
from pos_model import OrderParser
from pos_util import SaleLineUtil
from systools import sys_parsefilepath
from typing import (
    Tuple,
    Union,
    Dict,
    List,
    Any,
)

import nfcebuilder
import satbuilder
from aposprinter import SatAposA8FiscalPrinter, NfceAposA8FiscalPrinter
from clichehelper import ClicheHelper
from common import (
    FiscalParameterController,
    TipBuilderCommon,
    PrinterCommon,
    CalculateTaxesItemCommon,
)
from eventhandler import FiscalWrapperEventHandler
from ibpttaxprocessor import IbptTaxProcessor
from models import (
    SoftwareDeveloper,
    LoyaltyConfigurations,
)
from nfce import (
    NfceProcessor,
    NfceRequestBuilder,
    NfceContingencia,
    NfceAutorizador,
    NfceConnectivityTester,
    NfceSituationChecker,
)
from nfcebuilder import NfeBuilder
from nfcebuilder.nfceutil import NfeSigner, XmlEnveloper, NfceRequest
from nfcecanceler import (
    NfceCanceler,
    NfceCancelerRequestBuilder,
    NfceCancelerParameters,
    NfceCancelerResponseParser,
)
from nfcedisabler import (
    OrderDisabler,
    NfceDisabler,
    NfceDisablerRequestBuilder,
    NfceDisablerParameters,
    NfceDisablerResponseParser,
    OrderRepository,
    ProcInutNfeBuilder,
    ProcInutNfeSaver,
)
from order_retriever import OrderRetrieverService, OrderDataFormatter
from order_retriever.repository import OrderRetrieverRepository
from printer import SatFiscalPrinter, NfceFiscalPrinter, DonationCouponPrinter
from repository import FiscalParameterRepository, SequencerRepository, ProductRepository
from satbuilder import CfeBuilder
from services import (
    FeeService,
    SequencerService,
)
from sat import (
    SatProcessor,
    SatRequestBuilder,
    SatServiceFinder,
)
from operator import itemgetter

REQUIRED_SERVICES = "FiscalPersistence|PosController|StoreWideConfig"


def main():
    loader_cfg = os.environ["LOADERCFG"]
    import_pydevd(loader_cfg, 9125)

    config = cfgtools.read(loader_cfg)
    service_name = str(config.find_value("Identification.ComponentName"))
    logger = logging.getLogger(service_name)

    config_logger(loader_cfg, service_name)
    config_logger(loader_cfg, service_name + "Thread")
    config_logger(loader_cfg, service_name + "NfceDisabler")
    config_logger(loader_cfg, service_name + "NfceCanceler")
    config_logger(loader_cfg, service_name + "Coupons")
    config_logger(
        loader_path=loader_cfg,
        log_name=service_name + "XmlLog",
    )

    xml_enviados = str(config.find_value("FiscalWrapper.XmlEnviados"))
    xml_erros = str(config.find_value("FiscalWrapper.XmlErros"))
    xml_inconsistentes = str(config.find_value("FiscalWrapper.XmlInconsistentes") or "Inconsistentes")
    export_xml_path = os.path.join(sys_parsefilepath(str(config.find_value("FiscalWrapper.XmlPath"))), service_name)
    crt = int(config.find_value("FiscalWrapper.crt"))
    ibpt_tax_files_path = sys_parsefilepath(str(config.find_value(
        "FiscalWrapper.IbptTaxFilesPath") or "../data/bundles/fiscalwrapper/ibpttaxfiles"))

    mbcontext = MBEasyContext(service_name)
    pyscripts.mbcontext = mbcontext
    sysactions.mbcontext = mbcontext

    nf_type = config.find_value("FiscalWrapper.NFType")
    order_ref_ids = _get_order_ref_ids(config)
    order_ref_position = (config.find_value("FiscalWrapper.OrderReferencePosition") or "BETWEEN_FISCAL_DATA").upper()

    fiscal_printer = None
    mobile_fiscal_printer = None
    fiscal_processor = None

    fiscal_sent_dir = sys_parsefilepath(config.find_value("FiscalWrapper.FiscalSentDir").strip())

    message_handler = MessageHandler(mbcontext, service_name, service_name, REQUIRED_SERVICES)

    nome_emit = remove_accents(read_swconfig(mbcontext, "Store.RazaoSocial"))
    end_logradouro = _limit_max_length(remove_accents(read_swconfig(mbcontext, "Store.EndLogradouro")), 60)
    end_numero = remove_punctuation(read_swconfig(mbcontext, "Store.EndNumero").strip()) or "S/N"
    end_compl = remove_accents((read_swconfig(mbcontext, "Store.EndCompl") or "").strip())
    bairro = remove_accents(read_swconfig(mbcontext, "Store.Bairro"))
    municipio = remove_accents(read_swconfig(mbcontext, "Store.Municipio"))
    cep = remove_punctuation(read_swconfig(mbcontext, "Store.CEP"))
    uf = remove_accents(read_swconfig(mbcontext, "Store.UF"))
    cnpj_contribuinte = remove_punctuation(config.find_value("FiscalWrapper.CnpjContribuinte"))
    inscr_estadual = remove_punctuation(config.find_value("FiscalWrapper.InscEstadual"))

    store_model = Store(
        crt,
        nome_emit,
        cnpj_contribuinte,
        inscr_estadual,
        cep,
        end_logradouro,
        end_numero,
        end_compl,
        bairro,
        municipio,
        uf
    )

    cliche_helper = ClicheHelper(mbcontext, cnpj_contribuinte, inscr_estadual)

    order_disabler = None
    nfce_contingencia = None
    nfce_situation_checker = None
    nfce_connectivity_tester = None
    nfce_canceler = None
    versao_ws = None
    synchronous_mode = False

    loader_fiscal_parameters = _get_config_group(
        config=config,
        group_path="FiscalWrapper.FiscalParameters",
    )
    fees_fiscal_parameters = _get_config_group(
        config=config,
        group_path="FiscalWrapper.FeesFiscalParameters",
    )
    loader_fiscal_parameters.extend(fees_fiscal_parameters)

    fiscal_coupon_configurations = _get_fiscal_coupon_configurations(config=config)
    fiscal_parameters = FiscalParameterRepository(mbcontext).get_fiscal_parameters(loader_fiscal_parameters)
    tenders_type = ProductRepository(mbcontext).get_tender_types()

    customer_text_field = config.find_value("FiscalWrapper.PrinterCustomizations.CustomerTextField") or ""
    print_order_identifier = (config.find_value(
        "FiscalWrapper.PrinterCustomizations.PrintOrderIdentifier") or "false").lower() == "true"
    customer_call_by_field = config.find_value("FiscalWrapper.PrinterCustomizations.YouWillBeCalledBy") or ""

    period_to_retry_orders_with_exception = config.find_value("FiscalWrapper.PeriodToRetryOrdersWithException") or "7"

    data_path = sys_parsefilepath("{HVDATADIR}")
    database_dir = os.path.join(data_path, "databases")

    default_fiscal_path = "../../src/persistcomp/fiscal/fiscal.ddl"
    fiscal_ddl_path = config.find_value("FiscalWrapper.FiscalSchemaDatabasePath", default=default_fiscal_path)
    fiscal_schema_sql_path = os.path.join(data_path, fiscal_ddl_path)

    order_parser = OrderParser()

    tip_config = _get_tip_config(config)
    tip_config_range = _get_tip_config_by_range(
        config=config,
        fiscal_parameters=fiscal_parameters,
    )
    calculate_taxes_common = CalculateTaxesItemCommon(mbcontext)
    fiscal_max_value = float(config.find_value("FiscalWrapper.MaxValueToFiscalize") or "10000")
    fiscal_parameter_controller = FiscalParameterController(
        fiscal_parameters=fiscal_parameters,
        tip_config_range=tip_config_range,
        loader_fiscal_parameters=loader_fiscal_parameters,
        fiscal_max_value=fiscal_max_value,
    )
    tip_builder_common = TipBuilderCommon(
        tip_config,
        tip_config_range,
        mbcontext,
        fiscal_parameter_controller,
        calculate_taxes_common
    )
    fee_service = FeeService(
        loader_config=config,
        logger=logger,
        calculate_taxes_common=calculate_taxes_common,
        fiscal_parameter_controller=fiscal_parameter_controller,
    )
    printer_common = PrinterCommon(
        tip_config=tip_config,
        tip_config_range=tip_config_range,
        tip_builder_common=tip_builder_common,
        customer_call_by_field=customer_call_by_field,
    )
    loyalty_configurations = _get_loyalty_configurations(
        mb_context=mbcontext,
        config=config,
    )

    ibtp_tax_processor = IbptTaxProcessor(mbcontext, ibpt_tax_files_path, uf, fiscal_parameter_controller)

    sale_line_util = SaleLineUtil()

    order_retriever_repository = OrderRetrieverRepository(mbcontext)
    order_data_formatter = OrderDataFormatter()
    order_retriever_service = OrderRetrieverService(order_retriever_repository, order_data_formatter)

    initial_order_id = None
    font_size_withdrawal_panel = _get_font_size_withdrawal_panel_mobile(
        config=config,
    )

    danfe_font_size_mobile = _get_danfe_font_size_mobile(
        config=config,
    )

    if nf_type.upper() == "SAT" or nf_type.upper() == "MFE":
        sat_group = config.find_group("FiscalWrapper.SAT")
        cnpj_sw = sat_group.find_value("CnpjSW")
        sign_ac = sat_group.find_value("SignAC")
        max_sat_number = int(sat_group.find_value("MaxSatNumber"))
        pos_sat_list = sat_group.find_value("PosSatList")
        sleep_time = int(sat_group.find_value("SearchSatInterval"))
        layout_version = sat_group.find_value("LayoutVersion", default="0.07")

        sat_service_finder = SatServiceFinder(
            mbcontext=mbcontext,
            max_sat_number=max_sat_number,
            pos_sat_list=pos_sat_list,
            sleep_time=sleep_time,
        )

        cfe_builder = _build_sat_builders(
            crt=crt,
            cnpj_contribuinte=cnpj_contribuinte,
            inscr_estadual=inscr_estadual,
            cnpj_sw=cnpj_sw,
            sign_ac=sign_ac,
            fiscal_parameter_controller=fiscal_parameter_controller,
            sale_line_util=sale_line_util,
            tip_config=tip_config,
            ibtp_tax_processor=ibtp_tax_processor,
            customer_text_field=customer_text_field,
            tenders_type=tenders_type,
            tip_builder_common=tip_builder_common,
            layout_version=layout_version,
            fee_service=fee_service,
        )

        sat_request_builder = SatRequestBuilder(
            mbcontext=mbcontext,
            crt=crt,
            cnpj_sw=cnpj_sw,
            sign_ac=sign_ac,
            cnpj_contribuinte=cnpj_contribuinte,
            inscr_estadual=inscr_estadual,
            order_parser=order_parser,
            fiscal_parameter_controller=fiscal_parameter_controller,
            sale_line_util=sale_line_util,
            cfe_builder=cfe_builder,
        )
        fiscal_processor = SatProcessor(mbcontext, sat_service_finder, sat_request_builder, fiscal_sent_dir)
        fiscal_printer = SatFiscalPrinter(
            mb_context=mbcontext,
            order_ref_ids=order_ref_ids,
            cliche_helper=cliche_helper,
            customer_text_field=customer_text_field,
            print_order_identifier=print_order_identifier,
            printer_common=printer_common,
            coupon_fiscal_parameters=fiscal_coupon_configurations,
            order_ref_position=order_ref_position,
            loyalty_configurations=loyalty_configurations,
        )
        mobile_fiscal_printer = SatAposA8FiscalPrinter(
            mb_context=mbcontext,
            order_ref_ids=order_ref_ids,
            cliche_helper=cliche_helper,
            customer_text_field=customer_text_field,
            printer_common=printer_common,
            coupon_fiscal_parameters=fiscal_coupon_configurations,
            font_size_withdrawal_panel=font_size_withdrawal_panel,
            danfe_font_size_mobile=danfe_font_size_mobile,
            order_ref_position=order_ref_position,
            loyalty_configurations=loyalty_configurations,
        )

    elif nf_type.upper() in ("NFCE"):
        nfce_group = config.find_group("FiscalWrapper.NFCE")

        c_uf = nfce_group.find_value("cUF")
        c_mun_fg = nfce_group.find_value("cMunFG")
        mod = nfce_group.find_value("mod")
        serie = nfce_group.find_value("serie")
        ambiente = nfce_group.find_value("ambiente")
        csc = nfce_group.find_value("csc")
        cid_token = nfce_group.find_value("cid_token")
        initial_order_id = int(nfce_group.find_value("initial_order_id"))

        serie_inc, initial_order_id_inc, dev_ambiente = _get_development_serie_and_order_id_increment()
        if serie_inc is not None:
            serie = str(int(serie) + serie_inc)
            initial_order_id += initial_order_id_inc
            ambiente = dev_ambiente

        versao_ws = int(nfce_group.find_value("versao_ws"))
        send_sleep_time = int(nfce_group.find_value("send_sleep_time") or 5)
        check_situation_timeout = int(nfce_group.find_value("CheckSituationTimeout") or 10)
        synchronous_mode = (nfce_group.find_value("synchronous_mode") or "false").lower() == "true"

        store_model.city_code = c_mun_fg
        store_model.state_code = c_uf

        software_developer = _read_software_developer_informations(config, uf)

        bundle_dir = os.environ["BUNDLEDIR"]
        certificate_key_path = bundle_dir + nfce_group.find_value("certificate_key_path")
        certificate_path = bundle_dir + nfce_group.find_value("certificate_path")
        nfe_signer = NfeSigner(certificate_key_path, certificate_path)
        sefaz_certificate_path = bundle_dir + nfce_group.find_value("sefaz_certificate_path")

        nfce_urls = _get_nfce_urls(logger, nfce_group, uf, versao_ws, ambiente)

        nfe_builder = _build_nfe_builders(
            serie=serie,
            mod=mod,
            ambiente=ambiente,
            fiscal_parameter_controller=fiscal_parameter_controller,
            sale_line_util=sale_line_util,
            cid_token=cid_token,
            csc=csc,
            qrcode_base_url=nfce_urls["qrcode_base_url"],
            nfe_signer=nfe_signer,
            versao_ws=versao_ws,
            qrcode_check_url=nfce_urls["qrcode_check_url"],
            software_developer=software_developer,
            ibtp_tax_processor=ibtp_tax_processor,
            customer_text_field=customer_text_field,
            tenders_type=tenders_type,
            tip_builder_common=tip_builder_common,
            store_model=store_model,
            fee_service=fee_service,
        )

        nfce_request_builder = NfceRequestBuilder(
            mbcontext,
            initial_order_id,
            crt,
            cnpj_contribuinte,
            inscr_estadual,
            c_uf,
            uf,
            c_mun_fg,
            mod,
            serie,
            ambiente,
            end_logradouro,
            end_numero,
            end_compl,
            bairro,
            municipio,
            cep,
            nome_emit,
            csc,
            cid_token,
            nfce_urls["qrcode_base_url"],
            certificate_key_path,
            certificate_path,
            order_parser,
            fiscal_parameter_controller,
            sale_line_util,
            nfe_builder,
            versao_ws,
            synchronous_mode
        )
        nfce_request = NfceRequest(
            certificate_key_path,
            certificate_path,
            sefaz_certificate_path,
            check_situation_timeout
        )

        max_tantavivas_envio_lote = int(nfce_group.find_value("max_tantavivas_envio_lote") or "3")
        max_tantavivas_envio_contingencia = int(nfce_group.find_value("max_tantavivas_envio_contingencia") or "10")
        intervalo_retentativa_lote = int(nfce_group.find_value("intervalo_retentativa_lote") or "15")

        if nf_type.upper() == "NFCE":
            nfce_connectivity_tester = NfceConnectivityTester(
                nfce_request, nfce_urls["url_status_servico"], c_uf, versao_ws)
            nfce_situation_checker = NfceSituationChecker(
                nfce_request, nfce_urls["url_consulta_nfe"], ambiente, c_uf, versao_ws)
            nfce_autorizador = NfceAutorizador(
                nfce_request_builder,
                nfce_request,
                nfce_urls["url_autorizacao"],
                nfce_urls["url_ret_autorizacao"],
                versao_ws,
                max_tantavivas_envio_lote,
                intervalo_retentativa_lote,
                send_sleep_time,
                synchronous_mode
            )
            nfce_contingencia = NfceContingencia(
                mbcontext=mbcontext,
                nfce_request_builder=nfce_request_builder,
                nfce_autorizador=nfce_autorizador,
                versao_ws=versao_ws,
                nfce_situation_checker=nfce_situation_checker,
                fiscal_sent_dir=fiscal_sent_dir,
                max_tantavivas_envio_contingencia=max_tantavivas_envio_contingencia,
                synchronous_mode=synchronous_mode,
                nfce_connectivity_tester=nfce_connectivity_tester,
                database_dir=database_dir,
                fiscal_schema_sql_path=fiscal_schema_sql_path,
            )
            fiscal_printer = NfceFiscalPrinter(
                mb_context=mbcontext,
                order_ref_ids=order_ref_ids,
                cliche_helper=cliche_helper,
                qrcode_check_url=nfce_urls["qrcode_check_url"],
                versao_ws=versao_ws,
                customer_text_field=customer_text_field,
                print_order_identifier=print_order_identifier,
                printer_common=printer_common,
                coupon_fiscal_parameters=fiscal_coupon_configurations,
                order_ref_position=order_ref_position,
                loyalty_configurations=loyalty_configurations,
            )
            mobile_fiscal_printer = NfceAposA8FiscalPrinter(
                mb_context=mbcontext,
                order_ref_ids=order_ref_ids,
                cliche_helper=cliche_helper,
                qrcode_url=nfce_urls["qrcode_check_url"],
                customer_text_field=customer_text_field,
                printer_common=printer_common,
                coupon_fiscal_parameters=fiscal_coupon_configurations,
                font_size_withdrawal_panel=font_size_withdrawal_panel,
                danfe_font_size_mobile=danfe_font_size_mobile,
                order_ref_position=order_ref_position,
                loyalty_configurations=loyalty_configurations,
            )
        else:
            nfce_request = None
            nfce_autorizador = None
            nfce_contingencia = None
            fiscal_printer = None
            nfce_situation_checker = None

        fiscal_processor = NfceProcessor(
            mbcontext=mbcontext,
            nfce_request_builder=nfce_request_builder,
            nfce_request=nfce_request,
            nfce_autorizador=nfce_autorizador,
            nfce_contingencia=nfce_contingencia,
            url_sefaz=nfce_urls["qrcode_base_url"],
            fiscal_sent_dir=fiscal_sent_dir,
            versao_ws=versao_ws,
            synchronous_mode=synchronous_mode,
            nfce_situation_checker=nfce_situation_checker,
            xml_enviados=xml_enviados,
        )

        justificativa_inutilizacao = nfce_group.find_value("justificativa_inutilizacao")
        nfce_disabler_parameters = NfceDisablerParameters(
            initial_order_id, ambiente, c_uf, cnpj_contribuinte, serie, justificativa_inutilizacao)
        nfce_canceler_parameters = NfceCancelerParameters(ambiente, c_uf, cnpj_contribuinte)

        xml_enveloper = XmlEnveloper(c_uf)
        nfce_disabler_request_builder = NfceDisablerRequestBuilder(
            nfce_disabler_parameters, nfe_signer, xml_enveloper, versao_ws)
        nfce_canceler_request_builder = NfceCancelerRequestBuilder(nfce_canceler_parameters, nfe_signer, xml_enveloper)

        nfce_disabler_response_parser = NfceDisablerResponseParser(versao_ws)
        nfce_canceler_response_parser = NfceCancelerResponseParser()

        nfce_canceler = NfceCanceler(
            nfce_canceler_request_builder,
            nfce_canceler_response_parser,
            nfce_request,
            nfce_urls["url_cancelamento"],
            fiscal_sent_dir,
            versao_ws
        )

        proc_inut_nfe_builder = ProcInutNfeBuilder()
        proc_inut_nfe_saver = ProcInutNfeSaver(fiscal_sent_dir)
        nfce_disabler = NfceDisabler(
            nfce_disabler_request_builder,
            nfce_disabler_response_parser,
            nfce_request,
            nfce_urls["url_inutilizacao"],
            proc_inut_nfe_builder,
            proc_inut_nfe_saver,
            versao_ws
        )

        order_taker = OrderTaker()

        pos_list = PosUtil(mbcontext).get_pos_list()
        order_repository = OrderRepository(mbcontext, pos_list)

        order_disabler = OrderDisabler(order_repository, nfce_disabler, order_taker)

    sequencer_repository = SequencerRepository(mbcontext)
    sequencer_service = SequencerService(sequencer_repository)

    last_order_id = int(order_retriever_service.get_last_order_id())
    initial_order_id = initial_order_id or 0
    sequencer_service.update_fiscal_id_sequence(initial_order_id, last_order_id)

    printers = {
        "": fiscal_printer,
        "APOS_A8": mobile_fiscal_printer,
        "EXTERNAL": mobile_fiscal_printer,
    }
    donation_coupon_printer = DonationCouponPrinter(cliche_helper, nf_type.upper())

    event_handler = FiscalWrapperEventHandler(
        mbcontext=mbcontext,
        fiscal_processor=fiscal_processor,
        printers=printers,
        nfce_connectivity_tester=nfce_connectivity_tester,
        order_disabler=order_disabler,
        nfce_canceler=nfce_canceler,
        fiscal_parameter_controller=fiscal_parameter_controller,
        xml_enviados=xml_enviados,
        xml_erros=xml_erros,
        export_xml_path=export_xml_path,
        fiscal_sent_dir=fiscal_sent_dir,
        nf_type=nf_type,
        versao_ws=versao_ws,
        ibpt_tax_processor=ibtp_tax_processor,
        order_retriever_service=order_retriever_service,
        xml_inconsistentes=xml_inconsistentes,
        donation_coupon_printer=donation_coupon_printer,
        nfce_contingencia=nfce_contingencia,
        period_to_retry_orders_with_exception=period_to_retry_orders_with_exception,
        synchronous_mode=synchronous_mode,
        sequencer_service=sequencer_service,
    )
    message_handler.set_event_handler(event_handler)

    if nfce_contingencia is not None:
        nfce_contingencia.start_contingencia()

    message_handler.subscribe_events([
        "ORDER_MODIFIED",
        "DisableNfceOrder",
        "ReSignXMLs",
        "OrderManagerFiscalXmlRequest",
        "CheckIfFiscalXmlIsCreatedInFolder",
        "RetryFiscalOrdersWithException"
    ])

    message_handler.handle_events()


def _get_nfce_urls(logger, nfce_group, uf, versao_ws, ambiente):
    default_nfce_urls = "../data/server/bundles/fiscalwrapper/nfce_urls.xml"
    urls_path = sys_parsefilepath(nfce_group.find_value("urls_path")) or default_nfce_urls
    nfce_urls = {}

    if os.path.exists(urls_path):
        logger.info("Reading NFCe URLs from <nfce_urls.xml>...")
        urls_xml = eTree.parse(urls_path).getroot()
        for state in urls_xml.findall("estado"):
            if state.get("uf") == uf:
                for version in state.findall("versao"):
                    if int(version.get("ws")) == versao_ws and version.get("ambiente") == ambiente:
                        for url in version:
                            nfce_urls[url.tag] = str.strip(url.text)

    if not nfce_urls:
        logger.info("NFCe URLs not found in <nfce_urls.xml>, reading from <loader.cfg>...")
        nfce_urls["url_autorizacao"] = nfce_group.find_value("url_autorizacao") or ""
        nfce_urls["url_ret_autorizacao"] = nfce_group.find_value("url_ret_autorizacao") or ""
        nfce_urls["url_inutilizacao"] = nfce_group.find_value("url_inutilizacao") or ""
        nfce_urls["url_cancelamento"] = nfce_group.find_value("url_cancelamento") or ""
        nfce_urls["url_status_servico"] = nfce_group.find_value("url_status_servico") or ""
        nfce_urls["url_consulta_nfe"] = nfce_group.find_value("url_consulta_nfe") or ""
        nfce_urls["qrcode_base_url"] = nfce_group.find_value("qrcode_base_url") or ""
        nfce_urls["qrcode_check_url"] = nfce_group.find_value("qrcode_check_url") or ""

    for url in nfce_urls:
        if not nfce_urls[url]:
            logger.warning("NFCe URL missing: [{}]".format(url))

    return nfce_urls


def _get_config_group(config, group_path):
    group = config.find_group(group_path)
    configs = []
    if group:
        for group in group.groups:
            config = {group.descr: group.name}
            for key in group.keys:
                config[key.name] = key.values[0]
            configs.append(config)
    return configs


def _build_nfe_builders(
    serie,                          # type: int
    mod,                            # type: int
    ambiente,                       # type: str
    fiscal_parameter_controller,    # type: FiscalParameterController
    sale_line_util,                 # type: SaleLineUtil
    cid_token,                      # type: str
    csc,                            # type: str
    qrcode_base_url,                # type: str
    nfe_signer,                     # type: NfeSigner
    versao_ws,                      # type: int
    qrcode_check_url,               # type: str
    software_developer,             # type: SoftwareDeveloper
    ibtp_tax_processor,             # type: IbptTaxProcessor
    customer_text_field,            # type: str
    tenders_type,                   # type: Dict[str, Any]
    tip_builder_common,             # type: TipBuilderCommon
    store_model,                    # type: Store
    fee_service,                    # type: FeeService
):
    # type: (...) -> NfeBuilder

    ide_builder = nfcebuilder.IdeBuilder(serie, mod, ambiente, versao_ws, tenders_type, store_model)
    emit_builder = nfcebuilder.EmitBuilder(store_model)

    crt = store_model.crt
    dest_builder = nfcebuilder.DestinatarioBuilder(ambiente)
    inf_resp_tec_builder = nfcebuilder.InfRespTecBuilder(software_developer)
    det_builder = nfcebuilder.DetBuilder(ambiente, crt, fiscal_parameter_controller, sale_line_util, versao_ws)
    det_tip_builder = nfcebuilder.DetTipBuilder(det_builder, tip_builder_common)
    det_fee_builder = nfcebuilder.DetFeeBuilder(
        det_builder=det_builder,
        fee_service=fee_service,
    )
    det_list_builder = nfcebuilder.DetListBuilder(
        det_builder,
        det_tip_builder,
        det_fee_builder,
    )
    total_builder = nfcebuilder.TotalBuilder(versao_ws)
    transp_builder = nfcebuilder.TranspBuilder(store_model)
    inf_intermed_builder = nfcebuilder.InfIntermedBuilder()
    pag_builder = nfcebuilder.PagBuilder(versao_ws, tenders_type)
    adic_builder = nfcebuilder.InfAdicBuilder(customer_text_field, ibtp_tax_processor)

    builders = [ide_builder,
                emit_builder,
                dest_builder,
                det_list_builder,
                total_builder,
                transp_builder,
                pag_builder,
                inf_intermed_builder,
                adic_builder,
                inf_resp_tec_builder]

    c_uf = store_model.state_code
    cnpj_contribuinte = store_model.cnpj

    inf_nfe_builder = [nfcebuilder.InfNfeBuilder(builders, c_uf, cnpj_contribuinte, mod, serie, versao_ws)]
    nfe_supl_builder = nfcebuilder.InfNFeSuplBuilder(
        ambiente, cid_token, csc, qrcode_base_url, qrcode_check_url, versao_ws)
    nfe_builder = nfcebuilder.NfeBuilder(inf_nfe_builder, nfe_signer, nfe_supl_builder)
    return nfe_builder


def _build_sat_builders(
    crt,                            # type: int
    cnpj_contribuinte,              # type: str
    inscr_estadual,                 # type: str
    cnpj_sw,                        # type: str
    sign_ac,                        # type: str
    fiscal_parameter_controller,    # type: FiscalParameterController
    sale_line_util,                 # type: SaleLineUtil
    tip_config,                     # type: Dict[str, Any]
    ibtp_tax_processor,             # type: IbptTaxProcessor
    customer_text_field,            # type: str
    tenders_type,                   # type: Dict[str, Any]
    tip_builder_common,             # type: TipBuilderCommon
    layout_version,                 # type: str
    fee_service,                    # type: FeeService
):
    # type: (...) -> CfeBuilder

    ide_builder = satbuilder.IdeBuilder(cnpj_sw, sign_ac)
    emit_builder = satbuilder.EmitBuilder(cnpj_contribuinte, inscr_estadual)
    dest_builder = satbuilder.DestinatarioBuilder()
    det_builder = satbuilder.DetBuilder(crt, fiscal_parameter_controller, sale_line_util, ibtp_tax_processor)
    det_tip_builder = satbuilder.DetTipBuilder(det_builder, tip_config, tip_builder_common)
    det_fee_builder = satbuilder.DetFeeBuilder(
        det_builder=det_builder,
        fee_service=fee_service,
    )
    det_list_builder = satbuilder.DetListBuilder(
        det_builder,
        det_tip_builder,
        det_fee_builder,
    )
    total_builder = satbuilder.TotalBuilder(ibtp_tax_processor)
    pgto_builder = satbuilder.PgtoBuilder(tenders_type)
    adic_builder = satbuilder.InfAdicBuilder(customer_text_field, ibtp_tax_processor)
    builders = [ide_builder, emit_builder, dest_builder, det_list_builder, total_builder, pgto_builder, adic_builder]
    inf_cfe_builder = [
        satbuilder.InfCfeBuilder(
            part_builders=builders,
            layout_version=layout_version,
        )
    ]
    cfe_builder = satbuilder.CfeBuilder(inf_cfe_builder)
    return cfe_builder


def _read_software_developer_informations(config, uf):
    # type: (Configuration, unicode) -> Union[SoftwareDeveloper, None]
    sv_group = config.find_group("FiscalWrapper.SoftwareDeveloper")
    software_developer = SoftwareDeveloper()

    if sv_group:
        software_developer.cnpj = str(sv_group.find_value("CNPJ") or "")
        software_developer.contact = str(sv_group.find_value("Contact") or "")
        software_developer.email = str(sv_group.find_value("Email") or "")
        software_developer.phone = str(sv_group.find_value("Phone") or "")
        software_developer.csrt = str(sv_group.find_value("CSRT") or "")
        software_developer.id_csrt = str(sv_group.find_value("idCSRT") or "")

        enabled_states = sv_group.find_values("States")
        software_developer.enabled = True if uf in enabled_states else False

    return software_developer


def _limit_max_length(value, max_length):
    return value.strip()[:max_length]


def _get_tip_config(config):
    try:
        tip_name = config.find_value("FiscalWrapper.TipConfigs.TipName") or None
        tip_code = config.find_value("FiscalWrapper.TipConfigs.TipCode")
        if tip_name is None:
            return {}
        return {'TipName': tip_name, 'TipCode': tip_code}
    except (Exception,):
        return None


def _get_fiscal_coupon_configurations(
    config,     # type: eTree.XML
):
    # type: (...) -> Dict[str, bool]

    try:
        return dict(
            operator=(config.find_value("FiscalWrapper.CouponAttrs.OPERATOR") or "True").lower() == "true",
            order_or_table=(config.find_value("FiscalWrapper.CouponAttrs.ORDER_OR_TABLE") or "True").lower() == "true",
            pos_id=(config.find_value("FiscalWrapper.CouponAttrs.POS_ID") or "True").lower() == "true",
        )
    except (Exception,):
        return {}


def _get_tip_config_by_range(
    config,             # type: Configuration
    fiscal_parameters,  # type: Dict[Tuple[str, str], str]
):
    # type: (...) -> Union[dict, List[dict]]

    try:
        groups = _get_config_group(
            config=config,
            group_path="FiscalWrapper.TipConfigs.TipCode",
        )
        if len(groups) == 0:
            return {}

        configurations = {}
        for group in groups:
            configurations[str(group.get("id"))] = _get_configuration_tip(group, fiscal_parameters)

        return sorted(configurations.values(), key=itemgetter('PercentageUpper'))
    except (Exception,):
        return {}


def _get_configuration_tip(group, fiscal_parameters):
    # type: (Dict, Dict) -> Dict

    tip_name = str(group.get("TipName"))
    max_percentage = float(group.get("PercentageUpper") or 99999)
    tip_id = str(group.get("id"))
    key = (tip_id, "ICMS")
    exempts_code = (40, 41, 50, 60)
    exempt = True if key not in fiscal_parameters or fiscal_parameters[key] in exempts_code else False
    return {'TipName': tip_name, 'TipCode': tip_id, 'PercentageUpper': max_percentage, "exempt": exempt}


def _get_font_size_withdrawal_panel_mobile(
    config,     # type: Any
):
    # type: (...) -> str

    font_size = (config.find_value("FiscalWrapper.PrinterCustomizations.FontSizeWithdrawalPanelMobile") or "large")
    font_size = font_size.lower()

    if font_size in ['small', 'large']:
        return font_size

    return "large"


def _get_danfe_font_size_mobile(
    config,   # type: Any
):
    # type: (...) -> str

    font_size = (config.find_value("FiscalWrapper.PrinterCustomizations.DanfePOSMobileFontSize") or "small")
    return font_size.lower()


def _get_order_ref_ids(config):
    default_ref_ids = "PAGER_NUMBER|CUSTOMER_NAME|ORDER_ID"
    return config.find_value("FiscalWrapper.OrderReferenceIdentifier", default_ref_ids).upper().split("|")


def _get_development_serie_and_order_id_increment():
    if os.environ.get("DEBUG", "").lower() == "true":
        initial_order_id_inc_file_path = os.path.join(os.environ["BUNDLEDIR"], "initial_order_id.txt")
        if os.path.exists(initial_order_id_inc_file_path):
            with open(initial_order_id_inc_file_path, "rb") as f:
                file_content = f.read().decode("utf-8").strip()
                serie_inc, initial_order_id_inc = file_content.split(";")
                ambiente = "2"
                return int(serie_inc), int(initial_order_id_inc), ambiente

    return None, None, None


def _get_loyalty_configurations(
    mb_context,     # type: MBEasyContext
    config,         # type: eTree.XML
):
    # type: (...) -> LoyaltyConfigurations

    try:
        print_loyalty = config.find_value(
            "FiscalWrapper.Loyalty.PrintLoyaltyCouponAfterFiscalization"
        ) or "False"

        print_loyalty_delivery_order = config.find_value(
            "FiscalWrapper.Loyalty.PrintLoyaltyCouponForDeliveryOrder"
        ) or "False"

        text_to_print = config.find_value(
            "FiscalWrapper.Loyalty.TextToPrintInLoyaltyCoupon"
        ) or ""

        return LoyaltyConfigurations(
            print_loyalty_coupon_after_fiscalization=print_loyalty.lower() == "true",
            print_loyalty_coupon_delivery_orders=print_loyalty_delivery_order.lower() == "true",
            text_to_print_in_loyalty_coupon=text_to_print,
            store_name=read_swconfig(
                mbcontext=mb_context,
                key="Store.Name",
            ) or "",
        )
    except (Exception,):
        return LoyaltyConfigurations(
            print_loyalty_coupon_after_fiscalization=False,
            print_loyalty_coupon_delivery_orders=False,
            text_to_print_in_loyalty_coupon="",
            store_name="",
        )


if __name__ == "__main__":
    main()



import time

from logging import Logger

from ctypes import (
    c_long,
    create_string_buffer,
    addressof,
)
from threading import (
    Condition,
    Thread,
)

from systools import sys_log_exception
from typing import (
    Optional,
    Dict,
)

available_sitefs = {}
configured_sitefs = {}

available_sitef_lock = Condition()

search_thread_condition = Condition()
search_thread_running = False
search_thread = None


class SitefServiceFinder(object):
    def __init__(
        self,
        connection_test_interval,   # type: int
        sitef_ip_list,              # type: []
        store_id,                   # type: str
        pdv_tef,                    # type: int
        logger,                     # type: Logger
        dll_path,                   # type: str
    ):
        # type: (...) -> None

        from ctypes import WinDLL

        self.logger = logger
        self.pdv_tef = pdv_tef
        self.store_id = store_id
        self.connection_test_interval = float(connection_test_interval)
        self.sitef_ip_list = sitef_ip_list
        self.dll_path = dll_path

        self.sitef_dll = WinDLL(dll_path)
        self.configura_sitef_interativo = self.sitef_dll["ConfiguraIntSiTefInterativo"]
        self.inicia_sitef_interativo = self.sitef_dll["IniciaFuncaoSiTefInterativo"]
        self.continua_sitef_interativo = self.sitef_dll["ContinuaFuncaoSiTefInterativo"]
        self.finaliza_sitef_interativo = self.sitef_dll["FinalizaFuncaoSiTefInterativo"]

        self._set_available_ips(
            sitef_ip_list=self.sitef_ip_list,
        )
        self.start_search_thread()

    def _set_available_ips(
        self,
        sitef_ip_list,  # type: []
    ):
        global configured_sitefs

        for ip in sitef_ip_list:
            if ip not in configured_sitefs:
                configured_sitefs[ip] = []

            if self.pdv_tef not in configured_sitefs[ip]:
                configured_sitefs[ip].append(self.pdv_tef)

    def _should_not_check_ip(
        self,
    ):
        # type: (...) -> bool

        return self.connection_test_interval <= 0

    def find_sitef_service(
        self,
    ):
        # type: (...) -> Optional[Dict]

        global available_sitefs

        self.logger.debug("Starting => find_sitef_service")

        if len(available_sitefs) == 0:
            if self._should_not_check_ip():
                self.logger.debug("No IP found in list, reload IP list")
                for sitef_ip in self.sitef_ip_list[:]:
                    if sitef_ip not in available_sitefs:
                        available_sitefs[sitef_ip] = []

                    available_sitefs[sitef_ip].append(self.pdv_tef)

            if len(available_sitefs) == 0:
                self.logger.debug("No IP was found in the list, we return empty")
                return None

        response_ip = next((ip for ip in available_sitefs if self.pdv_tef in available_sitefs[ip]), None)
        self.logger.debug("Returning IP => {0}".format(response_ip))
        return response_ip

    def sitef_unavailable(
        self,
        ip_sitef,   # type: str
    ):
        # type: (...) -> None

        global available_sitefs

        self.logger.debug("Starting => sitef_unavailable")
        self.logger.debug("Removing IP from the list of valid IPs. IP => {0}".format(ip_sitef))

        if ip_sitef in available_sitefs:
            del available_sitefs[ip_sitef]

    @staticmethod
    def lock_sitef(

    ):
        # type: (...) -> None

        global available_sitef_lock

        available_sitef_lock.acquire()

    @staticmethod
    def release_sitef(

    ):
        # type: (...) -> None

        global available_sitef_lock

        available_sitef_lock.release()

    def wakeup_search(
        self,
    ):
        # type: (...) -> None

        self.logger.debug("Starting => wakeup_search")

        global search_thread_running, search_thread_condition

        if search_thread_running:
            with search_thread_condition:
                search_thread_condition.notifyAll()

        self.logger.debug("End => wakeup_search")

    def search_sitef_modules(
        self,
    ):
        # type: (...) -> None

        global configured_sitefs, available_sitefs

        self.logger.debug("===================================================================")
        self.logger.debug("Starting => search_sitef_modules")
        self.logger.debug("Number of IPs for validation => {0}".format(len(configured_sitefs)))

        for ip in configured_sitefs.keys():
            if ip in [None, 'None', '']:
                continue

            if ip in available_sitefs:
                available_sitefs[ip] = configured_sitefs[ip]
                self.logger.debug("IP | {0} | already validate".format(ip))
                continue

            self.logger.debug("Validating IP => {0}".format(ip))
            try:
                result = self.process(
                    pos_id=str(self.pdv_tef),
                    data_fiscal=time.strftime("%Y%m%d"),
                    hora_fiscal=time.strftime("%H%M%S"),
                    ip_sitef=ip,
                )
                if result:
                    self.logger.debug("Success in validating the IP => {0}".format(ip))
                    if ip not in available_sitefs:
                        available_sitefs[ip] = []
                    available_sitefs[ip].append(self.pdv_tef)
                    continue

                self.logger.warn("Failed to validate IP => {0}".format(ip))

                if ip in available_sitefs:
                    self.sitef_unavailable(
                        ip_sitef=ip,
                    )

            except Exception:
                self.logger.exception("Exception on validate IP => {0}".format(ip))

        self.logger.debug("Finished => search_sitef_modules")
        self.logger.debug("===================================================================")

    def start_search_thread(self):

        global search_thread_running, search_thread

        if not search_thread_running:
            search_thread_running = True
            if search_thread is None:
                self.logger.debug("Starting => start_search_thread")
                search_thread = Thread(target=self._search_thread_loop)
                search_thread.daemon = True
                search_thread.name = "search_sitef_modules"
                search_thread.start()

    def stop_search_thread(self):

        if self._should_not_check_ip():
            return

        global search_thread_running, search_thread_condition

        if search_thread_running:
            with search_thread_condition:
                search_thread_running = False
                search_thread_condition.notifyAll()

    def _search_thread_loop(self):

        if self._should_not_check_ip():
            return

        global search_thread_running, search_thread_condition

        while search_thread_running:
            self.search_sitef_modules()
            with search_thread_condition:
                search_thread_condition.wait(self.connection_test_interval)

    def process(
        self,           # type: SitefServiceFinder
        pos_id,         # type: str
        data_fiscal,    # type: str
        hora_fiscal,    # type: str
        ip_sitef=None,  # type: Optional[str]
    ):
        # type: (...) -> bool

        global available_sitef_lock

        with available_sitef_lock:
            return self._process(
                pos_id=pos_id,
                data_fiscal=data_fiscal,
                hora_fiscal=hora_fiscal,
                ip_sitef=ip_sitef
            )

    def _process(
        self,           # type: SitefServiceFinder
        pos_id,         # type: str
        data_fiscal,    # type: str
        hora_fiscal,    # type: str
        ip_sitef=None,  # type: Optional[str]
    ):
        # type: (...) -> bool

        self.logger.debug("===================================================================")
        self.logger.debug("Starting => TestSitefProcessor")

        if not ip_sitef:
            self.logger.debug("Call function find_sitef_service")
            ip_sitef = self.find_sitef_service()

        if not self._check_pin_pad():
            return False

        id_terminal = "SW99%04d" % int(pos_id)
        log_message = "TestSitefProcessor - IP Sitef: {0}, ID Loja: {1}, ID Terminal: {2}"
        self.logger.debug(log_message.format(ip_sitef, self.store_id, id_terminal))

        ret = -12
        while ret == -12:
            ret = self.configura_sitef_interativo(
                str(ip_sitef),
                str(self.store_id),
                id_terminal,
                0
            )

        if ret == 0:
            ret = self.inicia_sitef_interativo(
                111,
                "",
                '{}'.format(pos_id),
                data_fiscal,
                hora_fiscal,
                "",
                ""
            )
            if ret == 10000:
                comando = c_long(0)
                tipo_campo = c_long(0)
                tam_minimo = c_long(0)
                tam_maximo = c_long(0)
                tam_buffer = 32768
                output_buffer = create_string_buffer('\000' * tam_buffer)

                while ret == 10000:
                    ret = self.continua_sitef_interativo(
                        addressof(comando),
                        addressof(tipo_campo),
                        addressof(tam_minimo),
                        addressof(tam_maximo),
                        output_buffer,
                        tam_buffer,
                        0
                    )

                    if ret != 10000:
                        break

                    try:
                        if comando.value == 20:
                            output_buffer.value = "0"
                            time.sleep(0.5)

                    except Exception as ex:
                        self.logger.exception("Exception continua sitef interativo")
                        sys_log_exception("Exception Trapped on SiTef Processor %" + str(ex))

                response = "Teste Realizado com Sucesso" if ret == 0 else "Falha no Teste: Erro %s" % str(ret)
                self.logger.debug("TestSitefProcessor - Resposta: {}".format(response))

                self.finaliza_sitef_interativo(
                    "0",
                    '{}'.format(pos_id),
                    data_fiscal,
                    hora_fiscal,
                    None
                )

                self.logger.debug("Finished => Success TestSitefProcessor")
                self.logger.debug("===================================================================")

                return True if ret == 0 else False

        self.logger.debug("Finished => TestSitefProcessor - Erro Configurando SiTef Interativo")
        self.logger.debug("===================================================================")

        return False

    def _check_pin_pad(
        self,   # type: SitefServiceFinder
    ):
        # type: (...) -> bool

        self.logger.debug("check pin pad")

        time_limit = (time.time() + 5.0)
        while time.time() < time_limit:
            if self.sitef_dll["VerificaPresencaPinPad"]():
                return True
            time.sleep(0.2)

        return False



from messageprocessor import (
    MessageProcessorCallback,
    MessageProcessor,
)
from typing import (
    Any,
    List,
)

from ._ConnectionWrapper import ConnectionWrapper


class PersistTransactionCallback(MessageProcessorCallback):
    def __init__(
        self,
        conn,                                       # type: ConnectionWrapper
        do_not_need_transaction_processors=None,    # type: List[MessageProcessor]
    ):
        # type: (...) -> None

        self.conn = conn
        self.do_not_need_transaction = {}
        for processor in do_not_need_transaction_processors:
            self.do_not_need_transaction[processor.get_processor_name()] = True

    def process_begun(
        self,           # type: PersistTransactionCallback
        event_name,     # type: str
        request_id,     # type: str
        data,           # type: str
    ):
        # type: (...) -> None

        pass

    def input_parsed(
        self,           # type: PersistTransactionCallback
        event_name,     # type: str
        request_id,     # type: str
        input_obj,      # type: Any
    ):
        # type: (...) -> None

        if event_name not in self.do_not_need_transaction:
            self.conn.begin_transaction()

    def business_called(
        self,           # type: PersistTransactionCallback
        event_name,     # type: str
        request_id,     # type: str
        input_obj,      # type: Any
        output_obj,     # type: Any
    ):
        # type: (...) -> None

        if event_name not in self.do_not_need_transaction:
            self.conn.commit()

    def process_finished(
        self,           # type: PersistTransactionCallback
        event_name,     # type: str
        request_id,     # type: str
        input_obj,      # type: Any
        output_obj,     # type: Any
    ):
        # type: (...) -> None

        pass

    def parse_exception(
        self,           # type: PersistTransactionCallback
        event_name,     # type: str
        request_id,     # type: str
        data,           # type: Any
        exception,      # type: BaseException
    ):
        # type: (...) -> None

        pass

    def unhandled_exception(
        self,           # type: PersistTransactionCallback
        event_name,     # type: str
        request_id,     # type: str
        input_obj,      # type: Any
        exception,      # type: BaseException
    ):
        # type: (...) -> None

        if event_name not in self.do_not_need_transaction:
            self.conn.rollback()



"""Stubs for more_itertools.recipes"""
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
from typing_extensions import overload, Type

# Type and type variable definitions
_T = TypeVar('_T')
_U = TypeVar('_U')

def take(n: int, iterable: Iterable[_T]) -> List[_T]: ...
def tabulate(
    function: Callable[[int], _T], start: int = ...
) -> Iterator[_T]: ...
def tail(n: int, iterable: Iterable[_T]) -> Iterator[_T]: ...
def consume(iterator: Iterable[object], n: Optional[int] = ...) -> None: ...
@overload
def nth(iterable: Iterable[_T], n: int) -> Optional[_T]: ...
@overload
def nth(iterable: Iterable[_T], n: int, default: _U) -> Union[_T, _U]: ...
def all_equal(iterable: Iterable[object]) -> bool: ...
def quantify(
    iterable: Iterable[_T], pred: Callable[[_T], bool] = ...
) -> int: ...
def pad_none(iterable: Iterable[_T]) -> Iterator[Optional[_T]]: ...
def padnone(iterable: Iterable[_T]) -> Iterator[Optional[_T]]: ...
def ncycles(iterable: Iterable[_T], n: int) -> Iterator[_T]: ...
def dotproduct(vec1: Iterable[object], vec2: Iterable[object]) -> object: ...
def flatten(listOfLists: Iterable[Iterable[_T]]) -> Iterator[_T]: ...
def repeatfunc(
    func: Callable[..., _U], times: Optional[int] = ..., *args: Any
) -> Iterator[_U]: ...
def pairwise(iterable: Iterable[_T]) -> Iterator[Tuple[_T, _T]]: ...
@overload
def grouper(
    iterable: Iterable[_T], n: int
) -> Iterator[Tuple[Optional[_T], ...]]: ...
@overload
def grouper(
    iterable: Iterable[_T], n: int, fillvalue: _U
) -> Iterator[Tuple[Union[_T, _U], ...]]: ...
@overload
def grouper(  # Deprecated interface
    iterable: int, n: Iterable[_T]
) -> Iterator[Tuple[Optional[_T], ...]]: ...
@overload
def grouper(  # Deprecated interface
    iterable: int, n: Iterable[_T], fillvalue: _U
) -> Iterator[Tuple[Union[_T, _U], ...]]: ...
def roundrobin(*iterables: Iterable[_T]) -> Iterator[_T]: ...
def partition(
    pred: Optional[Callable[[_T], object]], iterable: Iterable[_T]
) -> Tuple[Iterator[_T], Iterator[_T]]: ...
def powerset(iterable: Iterable[_T]) -> Iterator[Tuple[_T, ...]]: ...
def unique_everseen(
    iterable: Iterable[_T], key: Optional[Callable[[_T], _U]] = ...
) -> Iterator[_T]: ...
def unique_justseen(
    iterable: Iterable[_T], key: Optional[Callable[[_T], object]] = ...
) -> Iterator[_T]: ...
@overload
def iter_except(
    func: Callable[[], _T], exception: Type[BaseException], first: None = ...
) -> Iterator[_T]: ...
@overload
def iter_except(
    func: Callable[[], _T],
    exception: Type[BaseException],
    first: Callable[[], _U],
) -> Iterator[Union[_T, _U]]: ...
@overload
def first_true(
    iterable: Iterable[_T], *, pred: Optional[Callable[[_T], object]] = ...
) -> Optional[_T]: ...
@overload
def first_true(
    iterable: Iterable[_T],
    default: _U,
    pred: Optional[Callable[[_T], object]] = ...,
) -> Union[_T, _U]: ...
def random_product(
    *args: Iterable[_T], repeat: int = ...
) -> Tuple[_T, ...]: ...
def random_permutation(
    iterable: Iterable[_T], r: Optional[int] = ...
) -> Tuple[_T, ...]: ...
def random_combination(iterable: Iterable[_T], r: int) -> Tuple[_T, ...]: ...
def random_combination_with_replacement(
    iterable: Iterable[_T], r: int
) -> Tuple[_T, ...]: ...
def nth_combination(
    iterable: Iterable[_T], r: int, index: int
) -> Tuple[_T, ...]: ...
def prepend(value: _T, iterator: Iterable[_U]) -> Iterator[Union[_T, _U]]: ...
def convolve(signal: Iterable[_T], kernel: Iterable[_T]) -> Iterator[_T]: ...



data = (
'jjael',    # 0x00
'jjaelg',    # 0x01
'jjaelm',    # 0x02
'jjaelb',    # 0x03
'jjaels',    # 0x04
'jjaelt',    # 0x05
'jjaelp',    # 0x06
'jjaelh',    # 0x07
'jjaem',    # 0x08
'jjaeb',    # 0x09
'jjaebs',    # 0x0a
'jjaes',    # 0x0b
'jjaess',    # 0x0c
'jjaeng',    # 0x0d
'jjaej',    # 0x0e
'jjaec',    # 0x0f
'jjaek',    # 0x10
'jjaet',    # 0x11
'jjaep',    # 0x12
'jjaeh',    # 0x13
'jjya',    # 0x14
'jjyag',    # 0x15
'jjyagg',    # 0x16
'jjyags',    # 0x17
'jjyan',    # 0x18
'jjyanj',    # 0x19
'jjyanh',    # 0x1a
'jjyad',    # 0x1b
'jjyal',    # 0x1c
'jjyalg',    # 0x1d
'jjyalm',    # 0x1e
'jjyalb',    # 0x1f
'jjyals',    # 0x20
'jjyalt',    # 0x21
'jjyalp',    # 0x22
'jjyalh',    # 0x23
'jjyam',    # 0x24
'jjyab',    # 0x25
'jjyabs',    # 0x26
'jjyas',    # 0x27
'jjyass',    # 0x28
'jjyang',    # 0x29
'jjyaj',    # 0x2a
'jjyac',    # 0x2b
'jjyak',    # 0x2c
'jjyat',    # 0x2d
'jjyap',    # 0x2e
'jjyah',    # 0x2f
'jjyae',    # 0x30
'jjyaeg',    # 0x31
'jjyaegg',    # 0x32
'jjyaegs',    # 0x33
'jjyaen',    # 0x34
'jjyaenj',    # 0x35
'jjyaenh',    # 0x36
'jjyaed',    # 0x37
'jjyael',    # 0x38
'jjyaelg',    # 0x39
'jjyaelm',    # 0x3a
'jjyaelb',    # 0x3b
'jjyaels',    # 0x3c
'jjyaelt',    # 0x3d
'jjyaelp',    # 0x3e
'jjyaelh',    # 0x3f
'jjyaem',    # 0x40
'jjyaeb',    # 0x41
'jjyaebs',    # 0x42
'jjyaes',    # 0x43
'jjyaess',    # 0x44
'jjyaeng',    # 0x45
'jjyaej',    # 0x46
'jjyaec',    # 0x47
'jjyaek',    # 0x48
'jjyaet',    # 0x49
'jjyaep',    # 0x4a
'jjyaeh',    # 0x4b
'jjeo',    # 0x4c
'jjeog',    # 0x4d
'jjeogg',    # 0x4e
'jjeogs',    # 0x4f
'jjeon',    # 0x50
'jjeonj',    # 0x51
'jjeonh',    # 0x52
'jjeod',    # 0x53
'jjeol',    # 0x54
'jjeolg',    # 0x55
'jjeolm',    # 0x56
'jjeolb',    # 0x57
'jjeols',    # 0x58
'jjeolt',    # 0x59
'jjeolp',    # 0x5a
'jjeolh',    # 0x5b
'jjeom',    # 0x5c
'jjeob',    # 0x5d
'jjeobs',    # 0x5e
'jjeos',    # 0x5f
'jjeoss',    # 0x60
'jjeong',    # 0x61
'jjeoj',    # 0x62
'jjeoc',    # 0x63
'jjeok',    # 0x64
'jjeot',    # 0x65
'jjeop',    # 0x66
'jjeoh',    # 0x67
'jje',    # 0x68
'jjeg',    # 0x69
'jjegg',    # 0x6a
'jjegs',    # 0x6b
'jjen',    # 0x6c
'jjenj',    # 0x6d
'jjenh',    # 0x6e
'jjed',    # 0x6f
'jjel',    # 0x70
'jjelg',    # 0x71
'jjelm',    # 0x72
'jjelb',    # 0x73
'jjels',    # 0x74
'jjelt',    # 0x75
'jjelp',    # 0x76
'jjelh',    # 0x77
'jjem',    # 0x78
'jjeb',    # 0x79
'jjebs',    # 0x7a
'jjes',    # 0x7b
'jjess',    # 0x7c
'jjeng',    # 0x7d
'jjej',    # 0x7e
'jjec',    # 0x7f
'jjek',    # 0x80
'jjet',    # 0x81
'jjep',    # 0x82
'jjeh',    # 0x83
'jjyeo',    # 0x84
'jjyeog',    # 0x85
'jjyeogg',    # 0x86
'jjyeogs',    # 0x87
'jjyeon',    # 0x88
'jjyeonj',    # 0x89
'jjyeonh',    # 0x8a
'jjyeod',    # 0x8b
'jjyeol',    # 0x8c
'jjyeolg',    # 0x8d
'jjyeolm',    # 0x8e
'jjyeolb',    # 0x8f
'jjyeols',    # 0x90
'jjyeolt',    # 0x91
'jjyeolp',    # 0x92
'jjyeolh',    # 0x93
'jjyeom',    # 0x94
'jjyeob',    # 0x95
'jjyeobs',    # 0x96
'jjyeos',    # 0x97
'jjyeoss',    # 0x98
'jjyeong',    # 0x99
'jjyeoj',    # 0x9a
'jjyeoc',    # 0x9b
'jjyeok',    # 0x9c
'jjyeot',    # 0x9d
'jjyeop',    # 0x9e
'jjyeoh',    # 0x9f
'jjye',    # 0xa0
'jjyeg',    # 0xa1
'jjyegg',    # 0xa2
'jjyegs',    # 0xa3
'jjyen',    # 0xa4
'jjyenj',    # 0xa5
'jjyenh',    # 0xa6
'jjyed',    # 0xa7
'jjyel',    # 0xa8
'jjyelg',    # 0xa9
'jjyelm',    # 0xaa
'jjyelb',    # 0xab
'jjyels',    # 0xac
'jjyelt',    # 0xad
'jjyelp',    # 0xae
'jjyelh',    # 0xaf
'jjyem',    # 0xb0
'jjyeb',    # 0xb1
'jjyebs',    # 0xb2
'jjyes',    # 0xb3
'jjyess',    # 0xb4
'jjyeng',    # 0xb5
'jjyej',    # 0xb6
'jjyec',    # 0xb7
'jjyek',    # 0xb8
'jjyet',    # 0xb9
'jjyep',    # 0xba
'jjyeh',    # 0xbb
'jjo',    # 0xbc
'jjog',    # 0xbd
'jjogg',    # 0xbe
'jjogs',    # 0xbf
'jjon',    # 0xc0
'jjonj',    # 0xc1
'jjonh',    # 0xc2
'jjod',    # 0xc3
'jjol',    # 0xc4
'jjolg',    # 0xc5
'jjolm',    # 0xc6
'jjolb',    # 0xc7
'jjols',    # 0xc8
'jjolt',    # 0xc9
'jjolp',    # 0xca
'jjolh',    # 0xcb
'jjom',    # 0xcc
'jjob',    # 0xcd
'jjobs',    # 0xce
'jjos',    # 0xcf
'jjoss',    # 0xd0
'jjong',    # 0xd1
'jjoj',    # 0xd2
'jjoc',    # 0xd3
'jjok',    # 0xd4
'jjot',    # 0xd5
'jjop',    # 0xd6
'jjoh',    # 0xd7
'jjwa',    # 0xd8
'jjwag',    # 0xd9
'jjwagg',    # 0xda
'jjwags',    # 0xdb
'jjwan',    # 0xdc
'jjwanj',    # 0xdd
'jjwanh',    # 0xde
'jjwad',    # 0xdf
'jjwal',    # 0xe0
'jjwalg',    # 0xe1
'jjwalm',    # 0xe2
'jjwalb',    # 0xe3
'jjwals',    # 0xe4
'jjwalt',    # 0xe5
'jjwalp',    # 0xe6
'jjwalh',    # 0xe7
'jjwam',    # 0xe8
'jjwab',    # 0xe9
'jjwabs',    # 0xea
'jjwas',    # 0xeb
'jjwass',    # 0xec
'jjwang',    # 0xed
'jjwaj',    # 0xee
'jjwac',    # 0xef
'jjwak',    # 0xf0
'jjwat',    # 0xf1
'jjwap',    # 0xf2
'jjwah',    # 0xf3
'jjwae',    # 0xf4
'jjwaeg',    # 0xf5
'jjwaegg',    # 0xf6
'jjwaegs',    # 0xf7
'jjwaen',    # 0xf8
'jjwaenj',    # 0xf9
'jjwaenh',    # 0xfa
'jjwaed',    # 0xfb
'jjwael',    # 0xfc
'jjwaelg',    # 0xfd
'jjwaelm',    # 0xfe
'jjwaelb',    # 0xff
)



import json
import logging
import time

from logging import Logger

from typing import Optional
from production.service import PrinterService

from messagebus import (
    DataType,
    DefaultToken,
    Message,
    MessageBus,
    TokenCreator,
    TokenPriority
)
from memorycache import (
    CacheManager,
    PeriodBasedExpiration,
)

PERIPHERALS_GROUP = "6"
MSGGRP_DEVICE_MANAGER = "D"
TK_PRN_PRINT = TokenCreator.create_token(TokenPriority.high, PERIPHERALS_GROUP, "101")
TK_GET_USED_SERVICE = TokenCreator.create_token(TokenPriority.low, MSGGRP_DEVICE_MANAGER, "1")

MAX_RETRY = 10

printer_logger = logging.getLogger("ProductionPrinter")


class PickListPrinter(object):
    def __init__(
        self,                           # type: PickListPrinter
        message_bus,                    # type: MessageBus
        timeout,                        # type: int
        name,                           # type: str
        logger,                         # type: Logger
        timeout_to_cleanup_devices,     # type: int
        kds_id="",                      # type: Optional[str]
        printer="",                     # type: Optional[str]
    ):
        # type: (...) -> None

        self.name = name
        self.logger = logger
        self.message_bus = message_bus
        self.timeout = timeout
        self.printer = printer
        self.cache_devices = CacheManager(
            cache_expiration=PeriodBasedExpiration(period_in_minutes=timeout_to_cleanup_devices),
            new_object_func=self._renew_cache_devices,
        )
        self.printer_service = PrinterService(
            message_bus=message_bus,
            timeout=self.timeout,
        )
        self.kds_id = kds_id

    def clear_printer_cache(
        self,   # type: PickListPrinter
    ):
        # type: (...) -> None

        self.cache_devices.force_renew_cache()

    def print_report(
        self,   # type: PickListPrinter
        text,   # type: str
    ):
        # type: (...) -> bool

        for attempt in range(1, MAX_RETRY):
            printer_name = self._get_available_printer()
            try:
                printer_logger.info("Sending to printer: {} - Attempt: {}".format(printer_name, attempt))

                if attempt == 1:
                    printer_logger.info("Printer: {} - \nPrinting:\n {}".format(printer_name, text))

                reply = self.printer_service.send_to_printer(
                    printer_name=printer_name,
                    text=text,
                )
                if reply.token != DefaultToken.TK_SYS_ACK:
                    log_message = "Invalid response from Printer: {} / {} - {}"
                    self.clear_printer_cache()
                    printer_logger.warning(log_message.format(printer_name, reply.token, reply.data))
                else:
                    printer_logger.info("Successfully printed on Printer: {}".format(printer_name))
                    return True
            except BaseException as _:
                if attempt >= MAX_RETRY - 1:
                    log_message = "Printer {} unavailable - Failed to Print Report"
                    printer_logger.warning(
                        log_message.format(printer_name),
                        exc_info=True,
                    )
                else:
                    log_message = "Error when trying to print on Printer {} in attempt {} - {}"
                    if printer_name is None and self.kds_id:
                        additional_info = "All Configured Printers for KDS ID {} are disabled".format(self.kds_id)
                    else:
                        additional_info = "Tried to print on printer {} but failed".format(printer_name)
                    printer_logger.info(
                        log_message.format(printer_name, attempt, additional_info),
                        exc_info=True,
                    )

            time.sleep(2 * attempt)

        return False

    def _renew_cache_devices(
        self,    # type: PickListPrinter
    ):
        # type: (...) -> str

        if not self.kds_id:
            return self.printer

        data = json.dumps({
            "kdsId": int(self.kds_id),
            "deviceType": "printer",
        })
        message = Message(
            token=TK_GET_USED_SERVICE,
            data_type=DataType.param,
            data=data,
            timeout=self.timeout,
        )
        try:
            response = self.message_bus.send_message(
                component='DeviceManager',
                message=message,
            )

            if response.token != DefaultToken.TK_SYS_ACK:
                self.logger.error("Printer not found, saving {} as KDS {} printer. Received message: {}".format(
                    self.printer,
                    self.kds_id,
                    response.data,
                ))
                updated_printer = self.printer
            else:
                updated_printer = response.data or self.printer
                self.logger.info("Saving {} as KDS {} printer".format(updated_printer, self.kds_id))
        except Exception as _:
            self.logger.error(
                "Error sending message for DeviceManager for KdsId {}".format(self.kds_id),
                exc_info=True,
            )
            return self.printer

        return updated_printer

    def _get_available_printer(
        self,   # type: PickListPrinter
    ):
        # type: (...) -> str

        return self.cache_devices.get_cached_object()



from abc import ABCMeta, abstractmethod


class Sleeper(object):
    __meta__ = ABCMeta

    @abstractmethod
    def sleep(self, secs):
        # type: (int) -> None
        raise NotImplementedError()



import sys

try:
    import chardet
except ImportError:
    import charset_normalizer as chardet
    import warnings

    warnings.filterwarnings('ignore', 'Trying to detect', module='charset_normalizer')

# This code exists for backwards compatibility reasons.
# I don't like it either. Just look the other way. :)

for package in ('urllib3', 'idna'):
    locals()[package] = __import__(package)
    # This traversal is apparently necessary such that the identities are
    # preserved (requests.packages.urllib3.* is urllib3.*)
    for mod in list(sys.modules):
        if mod == package or mod.startswith(package + '.'):
            sys.modules['requests.packages.' + mod] = sys.modules[mod]

target = chardet.__name__
for mod in list(sys.modules):
    if mod == target or mod.startswith(target + '.'):
        sys.modules['requests.packages.' + target.replace(target, 'chardet')] = sys.modules[mod]
# Kinda cool, though, right?



from abc import abstractmethod
from typing import Dict


class AsyncTargetSubjectIdFieldMapRepository(object):
    @abstractmethod
    def update_map(self, subject: str, id_field: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def load_map(self) -> Dict[str, str]:
        raise NotImplementedError()



# util/concurrency.py
# Copyright (C) 2005-2023 the SQLAlchemy authors and contributors
# <see AUTHORS file>
#
# This module is part of SQLAlchemy and is released under
# the MIT License: https://www.opensource.org/licenses/mit-license.php
# mypy: allow-untyped-defs, allow-untyped-calls

from __future__ import annotations

import asyncio  # noqa
import typing

have_greenlet = False
greenlet_error = None
try:
    import greenlet  # type: ignore # noqa: F401
except ImportError as e:
    greenlet_error = str(e)
    pass
else:
    have_greenlet = True
    from ._concurrency_py3k import await_only as await_only
    from ._concurrency_py3k import await_fallback as await_fallback
    from ._concurrency_py3k import greenlet_spawn as greenlet_spawn
    from ._concurrency_py3k import is_exit_exception as is_exit_exception
    from ._concurrency_py3k import AsyncAdaptedLock as AsyncAdaptedLock
    from ._concurrency_py3k import (
        _util_async_run as _util_async_run,
    )  # noqa: F401
    from ._concurrency_py3k import (
        _util_async_run_coroutine_function as _util_async_run_coroutine_function,  # noqa: F401, E501
    )

if not typing.TYPE_CHECKING and not have_greenlet:

    def _not_implemented():
        # this conditional is to prevent pylance from considering
        # greenlet_spawn() etc as "no return" and dimming out code below it
        if have_greenlet:
            return None

        raise ValueError(
            "the greenlet library is required to use this function."
            " %s" % greenlet_error
            if greenlet_error
            else ""
        )

    def is_exit_exception(e):  # noqa: F811
        return not isinstance(e, Exception)

    def await_only(thing):  # type: ignore  # noqa: F811
        _not_implemented()

    def await_fallback(thing):  # type: ignore  # noqa: F811
        return thing

    def greenlet_spawn(fn, *args, **kw):  # type: ignore  # noqa: F811
        _not_implemented()

    def AsyncAdaptedLock(*args, **kw):  # type: ignore  # noqa: F811
        _not_implemented()

    def _util_async_run(fn, *arg, **kw):  # type: ignore  # noqa: F811
        return fn(*arg, **kw)

    def _util_async_run_coroutine_function(fn, *arg, **kw):  # type: ignore  # noqa: F811,E501
        _not_implemented()



import time
from logging import Logger
from threading import Thread

from mw_helper import ensure_list
from production.model import (
    ProductionOrder,
    InvalidOrderToPrint,
)
from production.service import PrinterService
from production.util import parse_constants
from production.view import ProductionView
from production.view._OrderXml import OrderXml
from production.view.printerviewservices.reportgenerator import PickListReportGenerator
from production.view.printerviewservices.reportprinter import PickListPrinter
from typing import List


class PrinterView(ProductionView):
    """ class PrinterView(ProductionView)
    Production view used that formats a report (or slip) and prints it.
    """

    def __init__(
        self,
        name,                   # type: str
        message_bus,            # type: MessageBus
        report,                 # type: str
        printer,                # type: str
        picklist_name,          # type: str
        timeout,                # type: int
        production_manager,     # type: ProductionManager
        report_generator,       # type: PickListReportGenerator
        picklist_printer,       # type: PickListPrinter
        tags_on_print=None,     # type: List[str]
        logger=None,            # type: Logger
    ):
        # type: (...) -> None

        super(PrinterView, self).__init__(name, logger)
        self.report_generator = report_generator
        self.picklist_printer = picklist_printer

        self.message_bus = message_bus
        self.report = report
        self.printer = printer
        self.picklist_name = picklist_name
        self.timeout = timeout
        self.tags_on_print = ensure_list(tags_on_print)
        self.logger = logger
        self.order_xml = OrderXml()
        self.production_manager = production_manager
        self.printer_service = PrinterService(
            message_bus=self.message_bus,
            timeout=self.timeout,
        )

    def handle_order(self, order, synch=False):
        # type: (ProductionOrder, bool) -> Thread

        t = Thread(target=self.generate_and_print_report, args=[order])
        t.daemon = True
        t.start()
        return t

    def clear_printer_cache(
        self,   # type: PrinterView
    ):
        # type: (...) -> None

        self.picklist_printer.clear_printer_cache()

    def generate_and_print_report(
        self,   # type: PrinterView
        order,  # type: ProductionOrder
    ):
        # type: (...) -> None

        report = None
        is_reprint = self.production_manager.get_picklist_id(order=order) is not None
        for _ in range(1, 5):
            try:
                report = self.report_generator.production_pick_list(
                    production_order=order,
                    is_reprint=is_reprint,
                )
                break
            except InvalidOrderToPrint:
                break
            except (Exception, ):
                self.exception("Exception trying to generate picklist for Order {}", order.order_id)
                time.sleep(1)

        if report is None or report == "":
            return

        parsed_report = parse_constants(
            printer_text=report,
            pos_id=int(order.originator[-2:]),
            is_reprint=is_reprint,
        )
        print_ok = self.picklist_printer.print_report(text=parsed_report)
        self.production_manager.save_pick_list_order(
            order=order,
            receipt=report,
            status=print_ok,
        )

    def get_view_tags(self):
        return self.tags_on_print



from abc import ABCMeta
from logging import Logger

from typing import Optional

from dialogmanager import DialogManager
from messagebus import MessageBus

cache_last_receipt = {}
settings_mb_context = None


class Cache(object):
    __metaclass__ = ABCMeta

    def __init__(
        self,
        message_bus,        # type: MessageBus
        dialog_manager,     # type: DialogManager
        logger,             # type: Logger
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.dialog_manager = dialog_manager
        self.logger = logger

    @staticmethod
    def set_last_receipt(
        receipt_data,       # type: str
        eft_processor,      # type: str
    ):
        # type: (...) -> None

        if eft_processor in ['', None]:
            return

        global cache_last_receipt

        cache_last_receipt[str(eft_processor.upper())] = receipt_data

    @staticmethod
    def get_last_receipt(
        eft_processor,      # type: str
    ):
        # type: (...) -> Optional[str]

        if eft_processor in ['', None]:
            return

        global cache_last_receipt

        return cache_last_receipt.get(str(eft_processor.upper()), None)




# at_02.py - functions for handling AT-02 (SEPA Creditor identifier)
#
# Copyright (C) 2014-2016 Sergi Almacellas Abellana
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""SEPA Identifier of the Creditor (AT-02).

This identifier is indicated in the ISO 20022 data element `Creditor Scheme
Identification`. The creditor can be a legal entity, or an association that
is not a legal entity, or a person.

The first two digits contain the ISO country code, the next two are check
digits for the ISO 7064 Mod 97, 10 checksum, the next tree contain the
Creditor Business Code (or `ZZZ` if no business code used) and the remainder
contains the country-specific identifier.

>>> validate('ES 23 ZZZ 47690558N')
'ES23ZZZ47690558N'
>>> validate('ES2300047690558N')
'ES2300047690558N'
>>> compact('ES++()+23ZZZ4//7690558N')
'ES23ZZZ47690558N'
>>> calc_check_digits('ESXXZZZ47690558N')
'23'
"""

from stdnum.exceptions import *
from stdnum.iso7064 import mod_97_10
from stdnum.util import clean


# the valid characters we have
_alphabet = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'


def compact(number):
    """Convert the AT-02 number to the minimal representation. This strips
    the number of any valid separators and removes invalid characters."""
    return clean(number, ' -/?:().m\'+"').strip().upper()


def _to_base10(number):
    """Prepare the number to its base10 representation so it can be checked
    with the ISO 7064 Mod 97, 10 algorithm. That means excluding positions 5
    to 7 and moving the first four digits to the end."""
    return ''.join(str(_alphabet.index(x)) for x in number[7:] + number[:4])


def validate(number):
    """Check if the number provided is a valid AT-02."""
    number = compact(number)
    try:
        test_number = _to_base10(number)
    except Exception:  # noqa: B902
        raise InvalidFormat()
    # ensure that checksum is valid
    mod_97_10.validate(test_number)
    return number


def is_valid(number):
    """Check if the number provided is a valid AT-02."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def calc_check_digits(number):
    """Calculate the check digits that should be put in the number to make it
    valid. Check digits in the supplied number are ignored."""
    number = compact(number)
    # replace check digits with placeholders
    number = ''.join((number[:2], '00', number[4:]))
    return mod_97_10.calc_check_digits(_to_base10(number)[:-2])



# mako/_ast_util.py
# Copyright 2006-2023 the Mako authors and contributors <see AUTHORS file>
#
# This module is part of Mako and is released under
# the MIT License: http://www.opensource.org/licenses/mit-license.php

"""
    ast
    ~~~

    This is a stripped down version of Armin Ronacher's ast module.

    :copyright: Copyright 2008 by Armin Ronacher.
    :license: Python License.
"""


from _ast import Add
from _ast import And
from _ast import AST
from _ast import BitAnd
from _ast import BitOr
from _ast import BitXor
from _ast import Div
from _ast import Eq
from _ast import FloorDiv
from _ast import Gt
from _ast import GtE
from _ast import If
from _ast import In
from _ast import Invert
from _ast import Is
from _ast import IsNot
from _ast import LShift
from _ast import Lt
from _ast import LtE
from _ast import Mod
from _ast import Mult
from _ast import Name
from _ast import Not
from _ast import NotEq
from _ast import NotIn
from _ast import Or
from _ast import PyCF_ONLY_AST
from _ast import RShift
from _ast import Sub
from _ast import UAdd
from _ast import USub


BOOLOP_SYMBOLS = {And: "and", Or: "or"}

BINOP_SYMBOLS = {
    Add: "+",
    Sub: "-",
    Mult: "*",
    Div: "/",
    FloorDiv: "//",
    Mod: "%",
    LShift: "<<",
    RShift: ">>",
    BitOr: "|",
    BitAnd: "&",
    BitXor: "^",
}

CMPOP_SYMBOLS = {
    Eq: "==",
    Gt: ">",
    GtE: ">=",
    In: "in",
    Is: "is",
    IsNot: "is not",
    Lt: "<",
    LtE: "<=",
    NotEq: "!=",
    NotIn: "not in",
}

UNARYOP_SYMBOLS = {Invert: "~", Not: "not", UAdd: "+", USub: "-"}

ALL_SYMBOLS = {}
ALL_SYMBOLS.update(BOOLOP_SYMBOLS)
ALL_SYMBOLS.update(BINOP_SYMBOLS)
ALL_SYMBOLS.update(CMPOP_SYMBOLS)
ALL_SYMBOLS.update(UNARYOP_SYMBOLS)


def parse(expr, filename="<unknown>", mode="exec"):
    """Parse an expression into an AST node."""
    return compile(expr, filename, mode, PyCF_ONLY_AST)


def iter_fields(node):
    """Iterate over all fields of a node, only yielding existing fields."""

    for field in node._fields:
        try:
            yield field, getattr(node, field)
        except AttributeError:
            pass


class NodeVisitor:

    """
    Walks the abstract syntax tree and call visitor functions for every node
    found.  The visitor functions may return values which will be forwarded
    by the `visit` method.

    Per default the visitor functions for the nodes are ``'visit_'`` +
    class name of the node.  So a `TryFinally` node visit function would
    be `visit_TryFinally`.  This behavior can be changed by overriding
    the `get_visitor` function.  If no visitor function exists for a node
    (return value `None`) the `generic_visit` visitor is used instead.

    Don't use the `NodeVisitor` if you want to apply changes to nodes during
    traversing.  For this a special visitor exists (`NodeTransformer`) that
    allows modifications.
    """

    def get_visitor(self, node):
        """
        Return the visitor function for this node or `None` if no visitor
        exists for this node.  In that case the generic visit function is
        used instead.
        """
        method = "visit_" + node.__class__.__name__
        return getattr(self, method, None)

    def visit(self, node):
        """Visit a node."""
        f = self.get_visitor(node)
        if f is not None:
            return f(node)
        return self.generic_visit(node)

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node."""
        for field, value in iter_fields(node):
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, AST):
                        self.visit(item)
            elif isinstance(value, AST):
                self.visit(value)


class NodeTransformer(NodeVisitor):

    """
    Walks the abstract syntax tree and allows modifications of nodes.

    The `NodeTransformer` will walk the AST and use the return value of the
    visitor functions to replace or remove the old node.  If the return
    value of the visitor function is `None` the node will be removed
    from the previous location otherwise it's replaced with the return
    value.  The return value may be the original node in which case no
    replacement takes place.

    Here an example transformer that rewrites all `foo` to `data['foo']`::

        class RewriteName(NodeTransformer):

            def visit_Name(self, node):
                return copy_location(Subscript(
                    value=Name(id='data', ctx=Load()),
                    slice=Index(value=Str(s=node.id)),
                    ctx=node.ctx
                ), node)

    Keep in mind that if the node you're operating on has child nodes
    you must either transform the child nodes yourself or call the generic
    visit function for the node first.

    Nodes that were part of a collection of statements (that applies to
    all statement nodes) may also return a list of nodes rather than just
    a single node.

    Usually you use the transformer like this::

        node = YourTransformer().visit(node)
    """

    def generic_visit(self, node):
        for field, old_value in iter_fields(node):
            old_value = getattr(node, field, None)
            if isinstance(old_value, list):
                new_values = []
                for value in old_value:
                    if isinstance(value, AST):
                        value = self.visit(value)
                        if value is None:
                            continue
                        elif not isinstance(value, AST):
                            new_values.extend(value)
                            continue
                    new_values.append(value)
                old_value[:] = new_values
            elif isinstance(old_value, AST):
                new_node = self.visit(old_value)
                if new_node is None:
                    delattr(node, field)
                else:
                    setattr(node, field, new_node)
        return node


class SourceGenerator(NodeVisitor):

    """
    This visitor is able to transform a well formed syntax tree into python
    sourcecode.  For more details have a look at the docstring of the
    `node_to_source` function.
    """

    def __init__(self, indent_with):
        self.result = []
        self.indent_with = indent_with
        self.indentation = 0
        self.new_lines = 0

    def write(self, x):
        if self.new_lines:
            if self.result:
                self.result.append("\n" * self.new_lines)
            self.result.append(self.indent_with * self.indentation)
            self.new_lines = 0
        self.result.append(x)

    def newline(self, n=1):
        self.new_lines = max(self.new_lines, n)

    def body(self, statements):
        self.new_line = True
        self.indentation += 1
        for stmt in statements:
            self.visit(stmt)
        self.indentation -= 1

    def body_or_else(self, node):
        self.body(node.body)
        if node.orelse:
            self.newline()
            self.write("else:")
            self.body(node.orelse)

    def signature(self, node):
        want_comma = []

        def write_comma():
            if want_comma:
                self.write(", ")
            else:
                want_comma.append(True)

        padding = [None] * (len(node.args) - len(node.defaults))
        for arg, default in zip(node.args, padding + node.defaults):
            write_comma()
            self.visit(arg)
            if default is not None:
                self.write("=")
                self.visit(default)
        if node.vararg is not None:
            write_comma()
            self.write("*" + node.vararg.arg)
        if node.kwarg is not None:
            write_comma()
            self.write("**" + node.kwarg.arg)

    def decorators(self, node):
        for decorator in node.decorator_list:
            self.newline()
            self.write("@")
            self.visit(decorator)

    # Statements

    def visit_Assign(self, node):
        self.newline()
        for idx, target in enumerate(node.targets):
            if idx:
                self.write(", ")
            self.visit(target)
        self.write(" = ")
        self.visit(node.value)

    def visit_AugAssign(self, node):
        self.newline()
        self.visit(node.target)
        self.write(BINOP_SYMBOLS[type(node.op)] + "=")
        self.visit(node.value)

    def visit_ImportFrom(self, node):
        self.newline()
        self.write("from %s%s import " % ("." * node.level, node.module))
        for idx, item in enumerate(node.names):
            if idx:
                self.write(", ")
            self.write(item)

    def visit_Import(self, node):
        self.newline()
        for item in node.names:
            self.write("import ")
            self.visit(item)

    def visit_Expr(self, node):
        self.newline()
        self.generic_visit(node)

    def visit_FunctionDef(self, node):
        self.newline(n=2)
        self.decorators(node)
        self.newline()
        self.write("def %s(" % node.name)
        self.signature(node.args)
        self.write("):")
        self.body(node.body)

    def visit_ClassDef(self, node):
        have_args = []

        def paren_or_comma():
            if have_args:
                self.write(", ")
            else:
                have_args.append(True)
                self.write("(")

        self.newline(n=3)
        self.decorators(node)
        self.newline()
        self.write("class %s" % node.name)
        for base in node.bases:
            paren_or_comma()
            self.visit(base)
        # XXX: the if here is used to keep this module compatible
        #      with python 2.6.
        if hasattr(node, "keywords"):
            for keyword in node.keywords:
                paren_or_comma()
                self.write(keyword.arg + "=")
                self.visit(keyword.value)
            if getattr(node, "starargs", None):
                paren_or_comma()
                self.write("*")
                self.visit(node.starargs)
            if getattr(node, "kwargs", None):
                paren_or_comma()
                self.write("**")
                self.visit(node.kwargs)
        self.write(have_args and "):" or ":")
        self.body(node.body)

    def visit_If(self, node):
        self.newline()
        self.write("if ")
        self.visit(node.test)
        self.write(":")
        self.body(node.body)
        while True:
            else_ = node.orelse
            if len(else_) == 1 and isinstance(else_[0], If):
                node = else_[0]
                self.newline()
                self.write("elif ")
                self.visit(node.test)
                self.write(":")
                self.body(node.body)
            else:
                self.newline()
                self.write("else:")
                self.body(else_)
                break

    def visit_For(self, node):
        self.newline()
        self.write("for ")
        self.visit(node.target)
        self.write(" in ")
        self.visit(node.iter)
        self.write(":")
        self.body_or_else(node)

    def visit_While(self, node):
        self.newline()
        self.write("while ")
        self.visit(node.test)
        self.write(":")
        self.body_or_else(node)

    def visit_With(self, node):
        self.newline()
        self.write("with ")
        self.visit(node.context_expr)
        if node.optional_vars is not None:
            self.write(" as ")
            self.visit(node.optional_vars)
        self.write(":")
        self.body(node.body)

    def visit_Pass(self, node):
        self.newline()
        self.write("pass")

    def visit_Print(self, node):
        # XXX: python 2.6 only
        self.newline()
        self.write("print ")
        want_comma = False
        if node.dest is not None:
            self.write(" >> ")
            self.visit(node.dest)
            want_comma = True
        for value in node.values:
            if want_comma:
                self.write(", ")
            self.visit(value)
            want_comma = True
        if not node.nl:
            self.write(",")

    def visit_Delete(self, node):
        self.newline()
        self.write("del ")
        for idx, target in enumerate(node):
            if idx:
                self.write(", ")
            self.visit(target)

    def visit_TryExcept(self, node):
        self.newline()
        self.write("try:")
        self.body(node.body)
        for handler in node.handlers:
            self.visit(handler)

    def visit_TryFinally(self, node):
        self.newline()
        self.write("try:")
        self.body(node.body)
        self.newline()
        self.write("finally:")
        self.body(node.finalbody)

    def visit_Global(self, node):
        self.newline()
        self.write("global " + ", ".join(node.names))

    def visit_Nonlocal(self, node):
        self.newline()
        self.write("nonlocal " + ", ".join(node.names))

    def visit_Return(self, node):
        self.newline()
        self.write("return ")
        self.visit(node.value)

    def visit_Break(self, node):
        self.newline()
        self.write("break")

    def visit_Continue(self, node):
        self.newline()
        self.write("continue")

    def visit_Raise(self, node):
        # XXX: Python 2.6 / 3.0 compatibility
        self.newline()
        self.write("raise")
        if hasattr(node, "exc") and node.exc is not None:
            self.write(" ")
            self.visit(node.exc)
            if node.cause is not None:
                self.write(" from ")
                self.visit(node.cause)
        elif hasattr(node, "type") and node.type is not None:
            self.visit(node.type)
            if node.inst is not None:
                self.write(", ")
                self.visit(node.inst)
            if node.tback is not None:
                self.write(", ")
                self.visit(node.tback)

    # Expressions

    def visit_Attribute(self, node):
        self.visit(node.value)
        self.write("." + node.attr)

    def visit_Call(self, node):
        want_comma = []

        def write_comma():
            if want_comma:
                self.write(", ")
            else:
                want_comma.append(True)

        self.visit(node.func)
        self.write("(")
        for arg in node.args:
            write_comma()
            self.visit(arg)
        for keyword in node.keywords:
            write_comma()
            self.write(keyword.arg + "=")
            self.visit(keyword.value)
        if getattr(node, "starargs", None):
            write_comma()
            self.write("*")
            self.visit(node.starargs)
        if getattr(node, "kwargs", None):
            write_comma()
            self.write("**")
            self.visit(node.kwargs)
        self.write(")")

    def visit_Name(self, node):
        self.write(node.id)

    def visit_NameConstant(self, node):
        self.write(str(node.value))

    def visit_arg(self, node):
        self.write(node.arg)

    def visit_Str(self, node):
        self.write(repr(node.s))

    def visit_Bytes(self, node):
        self.write(repr(node.s))

    def visit_Num(self, node):
        self.write(repr(node.n))

    # newly needed in Python 3.8
    def visit_Constant(self, node):
        self.write(repr(node.value))

    def visit_Tuple(self, node):
        self.write("(")
        idx = -1
        for idx, item in enumerate(node.elts):
            if idx:
                self.write(", ")
            self.visit(item)
        self.write(idx and ")" or ",)")

    def sequence_visit(left, right):
        def visit(self, node):
            self.write(left)
            for idx, item in enumerate(node.elts):
                if idx:
                    self.write(", ")
                self.visit(item)
            self.write(right)

        return visit

    visit_List = sequence_visit("[", "]")
    visit_Set = sequence_visit("{", "}")
    del sequence_visit

    def visit_Dict(self, node):
        self.write("{")
        for idx, (key, value) in enumerate(zip(node.keys, node.values)):
            if idx:
                self.write(", ")
            self.visit(key)
            self.write(": ")
            self.visit(value)
        self.write("}")

    def visit_BinOp(self, node):
        self.write("(")
        self.visit(node.left)
        self.write(" %s " % BINOP_SYMBOLS[type(node.op)])
        self.visit(node.right)
        self.write(")")

    def visit_BoolOp(self, node):
        self.write("(")
        for idx, value in enumerate(node.values):
            if idx:
                self.write(" %s " % BOOLOP_SYMBOLS[type(node.op)])
            self.visit(value)
        self.write(")")

    def visit_Compare(self, node):
        self.write("(")
        self.visit(node.left)
        for op, right in zip(node.ops, node.comparators):
            self.write(" %s " % CMPOP_SYMBOLS[type(op)])
            self.visit(right)
        self.write(")")

    def visit_UnaryOp(self, node):
        self.write("(")
        op = UNARYOP_SYMBOLS[type(node.op)]
        self.write(op)
        if op == "not":
            self.write(" ")
        self.visit(node.operand)
        self.write(")")

    def visit_Subscript(self, node):
        self.visit(node.value)
        self.write("[")
        self.visit(node.slice)
        self.write("]")

    def visit_Slice(self, node):
        if node.lower is not None:
            self.visit(node.lower)
        self.write(":")
        if node.upper is not None:
            self.visit(node.upper)
        if node.step is not None:
            self.write(":")
            if not (isinstance(node.step, Name) and node.step.id == "None"):
                self.visit(node.step)

    def visit_ExtSlice(self, node):
        for idx, item in node.dims:
            if idx:
                self.write(", ")
            self.visit(item)

    def visit_Yield(self, node):
        self.write("yield ")
        self.visit(node.value)

    def visit_Lambda(self, node):
        self.write("lambda ")
        self.signature(node.args)
        self.write(": ")
        self.visit(node.body)

    def visit_Ellipsis(self, node):
        self.write("Ellipsis")

    def generator_visit(left, right):
        def visit(self, node):
            self.write(left)
            self.visit(node.elt)
            for comprehension in node.generators:
                self.visit(comprehension)
            self.write(right)

        return visit

    visit_ListComp = generator_visit("[", "]")
    visit_GeneratorExp = generator_visit("(", ")")
    visit_SetComp = generator_visit("{", "}")
    del generator_visit

    def visit_DictComp(self, node):
        self.write("{")
        self.visit(node.key)
        self.write(": ")
        self.visit(node.value)
        for comprehension in node.generators:
            self.visit(comprehension)
        self.write("}")

    def visit_IfExp(self, node):
        self.visit(node.body)
        self.write(" if ")
        self.visit(node.test)
        self.write(" else ")
        self.visit(node.orelse)

    def visit_Starred(self, node):
        self.write("*")
        self.visit(node.value)

    def visit_Repr(self, node):
        # XXX: python 2.6 only
        self.write("`")
        self.visit(node.value)
        self.write("`")

    # Helper Nodes

    def visit_alias(self, node):
        self.write(node.name)
        if node.asname is not None:
            self.write(" as " + node.asname)

    def visit_comprehension(self, node):
        self.write(" for ")
        self.visit(node.target)
        self.write(" in ")
        self.visit(node.iter)
        if node.ifs:
            for if_ in node.ifs:
                self.write(" if ")
                self.visit(if_)

    def visit_excepthandler(self, node):
        self.newline()
        self.write("except")
        if node.type is not None:
            self.write(" ")
            self.visit(node.type)
            if node.name is not None:
                self.write(" as ")
                self.visit(node.name)
        self.write(":")
        self.body(node.body)
